"""TLK modification algorithms for TSLPatcher/HoloPatcher.

This module implements TLK modification logic for applying patches from changes.ini files.
Handles string additions, modifications, and memory token resolution.

References:
----------

    Derivations and Other Implementations:
    -------------------------------------
        https://github.com/th3w1zard1/TSLPatcher/tree/master/TSLPatcher.pl (unfinished perl rewrite of TSLPatcher)
        https://github.com/th3w1zard1/HoloPatcher.NET/tree/master/src/TSLPatcher.Core/Mods/TLK/
        https://github.com/th3w1zard1/Kotor.NET/tree/master/Kotor.NET.Patcher/


"""

from __future__ import annotations

from typing import TYPE_CHECKING, ClassVar

from pykotor.common.misc import ResRef
from pykotor.extract.talktable import TalkTable
from pykotor.resource.formats.tlk.io_tlk import TLKBinaryReader
from pykotor.resource.formats.tlk.tlk_auto import bytes_tlk
from pykotor.tslpatcher.mods.template import PatcherModifications

if TYPE_CHECKING:
    from pathlib import Path

    from typing_extensions import Literal

    from pykotor.common.misc import Game
    from pykotor.resource.formats.tlk import TLK
    from pykotor.resource.type import SOURCE_TYPES
    from pykotor.tslpatcher.logger import PatchLogger
    from pykotor.tslpatcher.memory import PatcherMemory
    from utility.common.more_collections import CaseInsensitiveDict


class ModificationsTLK(PatcherModifications):
    DEFAULT_DESTINATION: ClassVar[str] = "."
    DEFAULT_SOURCEFILE: ClassVar[str] = "append.tlk"
    DEFAULT_SOURCEFILE_F: ClassVar[str] = "appendf.tlk"
    DEFAULT_SAVEAS_FILE: ClassVar[str] = "dialog.tlk"
    DEFAULT_SAVEAS_FILE_F: ClassVar[str] = "dialogf.tlk"

    def __init__(
        self,
        filename: str = DEFAULT_SOURCEFILE,
        *,
        replace: bool | None = None,  # noqa: FBT001, FBT002
        modifiers: list[ModifyTLK] | None = None,
    ):
        super().__init__(filename)
        self.destination: str = self.DEFAULT_DESTINATION
        self.modifiers: list[ModifyTLK] = [] if modifiers is None else modifiers
        self.sourcefile_f: str = self.DEFAULT_SOURCEFILE_F  # Polish version of k1
        self.saveas: str = self.DEFAULT_SAVEAS_FILE

    def pop_tslpatcher_vars(
        self,
        file_section_dict: CaseInsensitiveDict[str],
        default_destination: str | None = DEFAULT_DESTINATION,
        default_sourcefolder: str = ".",
    ):
        """Populates the TSLPatcher variables from the file section dictionary.

        Args:
        ----
            file_section_dict: CaseInsensitiveDict[str] - The file section dictionary
            default_destination: str | None - The default destination
            default_sourcefolder: str - The default source folder
        """
        if "!ReplaceFile" in file_section_dict:
            msg = "!ReplaceFile is not supported in [TLKList]"
            raise ValueError(msg)
        if "!OverrideType" in file_section_dict:
            msg = "!OverrideType is not supported in [TLKList]"
            raise ValueError(msg)

        self.sourcefile_f = file_section_dict.pop("!SourceFileF", self.DEFAULT_SOURCEFILE_F)
        super().pop_tslpatcher_vars(file_section_dict, default_destination, default_sourcefolder)

    def patch_resource(
        self,
        source: SOURCE_TYPES,
        memory: PatcherMemory,
        logger: PatchLogger,
        game: Game,
    ) -> bytes | Literal[True]:
        dialog: TLK = TLKBinaryReader(source).load()
        self.apply(dialog, memory, logger, game)
        return bytes_tlk(dialog)

    def apply(
        self,
        mutable_data: TLK,
        memory: PatcherMemory,
        logger: PatchLogger,
        game: Game,
    ):
        """Applies the TLK patches to the TLK.

        Args:
        ----
            mutable_data: TLK - The TLK to apply the patches to
            memory: PatcherMemory - The memory context
            logger: PatchLogger - The logger
            game: Game - The game
        """
        for modifier in self.modifiers:
            modifier.apply(mutable_data, memory)
            logger.complete_patch()


class ModifyTLK:
    def __init__(
        self,
        token_id: int,
        is_replacement: bool = False,  # noqa: FBT001, FBT002
    ):
        self.tlk_filepath: Path | None = None
        self.text: str = ""
        self.sound: ResRef = ResRef.from_blank()

        self.mod_index: int = token_id
        self.token_id: int = token_id
        self.is_replacement: bool = is_replacement

    def apply(
        self,
        dialog: TLK,
        memory: PatcherMemory,
    ):
        """Applies a TLK patch to a TLK.

        Args:
        ----
            dialog: TLK - The TLK to apply the patch to
            memory: PatcherMemory - The memory context

        Processing Logic:
        ----------------
            - Loads the TLK file
            - Replaces the token ID with the text and sound
            - Stores the new stringref in the memory context for append operations only.
        """
        self.load()
        if self.is_replacement:
            dialog.replace(self.token_id, self.text, str(self.sound))
        else:
            memory.memory_str[self.token_id] = dialog.add(self.text, str(self.sound))

    def load(self):
        """Loads the TLK file."""
        if self.tlk_filepath is None:
            return
        lookup_tlk = TalkTable(self.tlk_filepath)
        self.text = self.text or lookup_tlk.string(self.mod_index)
        self.sound = self.sound or lookup_tlk.sound(self.mod_index)



"""This module handles classes relating to editing WAV files.

WAV files in KotOR can be voice-over (VO) or sound effects (SFX) with various encodings.

KotOR Audio Format Reference:
----------------------------
    - SFX files (streammusic): Have 0xFFF360C4 header, skip 470 bytes, then RIFF/WAVE
    - VO files (streamvoice): Usually standard RIFF/WAVE with PCM or IMA ADPCM
    - Music with MP3: RIFF header with size=50, skip 58 bytes, then MP3 data

References:
----------
        Based on unified K1 (swkotor.exe) and TSL (swkotor2.exe) WAV/audio structure.
        Addresses: (K1: swkotor.exe, TSL: swkotor2.exe — verify/fill TSL via REVA when available).

        - ".wav" extension string: K1: 0x0074d308, TSL: TODO
        - "wav" resource type string: K1: 0x0074d32c, TSL: TODO
        - "wave" string (wave format identifier): K1: 0x0073f064, TSL: TODO
        - "RIFF" string: K1: 0x0074d324, TSL: TODO
        - "STREAMWAVES" string: K1: 0x0074df34, TSL: TODO
        - "HD0:STREAMWAVES\\%s" path format: K1: 0x0074a7f4, TSL: TODO
        - "HD0:STREAMMUSIC\\%s" path format: K1: 0x0074c304, TSL: TODO
        - ".\\streamwaves", "d:\\streamwaves" paths: K1: 0x0074df40, 0x0074df50, TSL: TODO
        Format notes: SFX 470-byte header (0xFF 0xF3 0x60 0xC4); MP3-in-WAV riffSize 50, skip 58 bytes.
        Derivations and Other Implementations:
        ----------
        https://github.com/th3w1zard1/KotOR.js/tree/master/src/audio/AudioFile.ts:9-162


"""

from __future__ import annotations
from pykotor.resource.formats._base import BiowareResource

from enum import IntEnum

from pykotor.resource.type import ResourceType


class WaveEncoding(IntEnum):
    """Wave encoding types used by Bioware.

        References:
        ----------
            Original BioWare engine binaries (TODO: Verify with REVA when available)

    Derivations and Other Implementations:
    -------------------------------------
        - https://github.com/th3w1zard1/KotOR.js/tree/master/src/enums/audio/AudioFileWaveEncoding.ts
    """

    PCM = 0x01  # Linear PCM (uncompressed)
    MS_ADPCM = 0x02  # Microsoft ADPCM
    ALAW = 0x06  # A-Law companded
    MULAW = 0x07  # μ-Law companded
    IMA_ADPCM = 0x11  # IMA ADPCM (also known as DVI ADPCM)
    MP3 = 0x55  # MPEG Layer 3


class AudioFormat(IntEnum):
    """Audio format types for the WAV wrapper."""

    WAVE = 1  # Standard RIFF/WAVE format
    MP3 = 2  # MP3 data (possibly wrapped in WAV)
    UNKNOWN = 0  # Unknown format


class WAVType(IntEnum):
    """The type of WAV file for KotOR obfuscation purposes."""

    VO = 1  # Voice over WAV (streamvoice, streamwaves)
    SFX = 2  # Sound effects WAV (streammusic/sounds with 470-byte header)


class WAV(BiowareResource):
    """Represents a WAV file.

    Attributes:
    ----------
        wav_type: The WAV type (VO or SFX)
        audio_format: The actual audio format (WAVE or MP3)
        encoding: The wave encoding type (for WAVE format)
        channels: Number of audio channels
        sample_rate: Audio sample rate
        bits_per_sample: Bits per audio sample
        block_align: Block alignment
        bytes_per_sec: Bytes per second
        data: The raw audio data (PCM samples for WAVE, raw bytes for MP3)

    Note:
    ----
        When audio_format is MP3, the data contains raw MP3 bytes that should be
        played directly without WAV parsing.
    """

    BINARY_TYPE = ResourceType.WAV

    def __init__(
        self,
        wav_type: WAVType = WAVType.VO,
        audio_format: AudioFormat = AudioFormat.WAVE,
        encoding: WaveEncoding | int = WaveEncoding.PCM,
        channels: int = 1,
        sample_rate: int = 44100,
        bits_per_sample: int = 16,
        bytes_per_sec: int = 0,
        block_align: int = 0,
        data: bytes | None = None,
    ):
        self.wav_type: WAVType = wav_type
        self.audio_format: AudioFormat = audio_format
        # Store encoding as int to allow unsupported values
        self.encoding: int = int(encoding) if isinstance(encoding, WaveEncoding) else encoding
        self.channels: int = channels
        self.sample_rate: int = sample_rate
        self.bits_per_sample: int = bits_per_sample
        self.bytes_per_sec: int = bytes_per_sec
        self.block_align: int = block_align or (channels * bits_per_sample // 8)
        self.data: bytes = data if data is not None else b""

    def get_encoding_enum(self) -> WaveEncoding | None:
        """Get the encoding as WaveEncoding enum, or None if unsupported."""
        try:
            return WaveEncoding(self.encoding)
        except ValueError:
            return None

    def is_pcm(self) -> bool:
        """Check if the audio is PCM encoded."""
        return self.encoding == WaveEncoding.PCM

    def is_adpcm(self) -> bool:
        """Check if the audio is IMA ADPCM encoded."""
        return self.encoding == WaveEncoding.IMA_ADPCM

    def is_mp3(self) -> bool:
        """Check if the audio is MP3 format."""
        return self.audio_format == AudioFormat.MP3 or self.encoding == WaveEncoding.MP3

    def __eq__(self, other):
        if self is other:
            return True
        if not isinstance(other, WAV):
            return NotImplemented  # type: ignore[no-any-return]
        return (
            self.wav_type == other.wav_type
            and self.audio_format == other.audio_format
            and self.encoding == other.encoding
            and self.channels == other.channels
            and self.sample_rate == other.sample_rate
            and self.bits_per_sample == other.bits_per_sample
            and self.block_align == other.block_align
            and self.data == other.data
        )

    def __hash__(self):
        return hash((self.wav_type, self.audio_format, self.encoding, self.channels, self.sample_rate, self.bits_per_sample, self.block_align, self.data))



"""Auto-detect and handle KotOR WAV files.

This module provides the public API for reading and writing WAV files.
It automatically handles KotOR's various audio obfuscation formats.

Usage:
------
    # Reading (automatic deobfuscation):
    wav = read_wav("path/to/file.wav")

    # Writing (with obfuscation for game compatibility):
    write_wav(wav, "output.wav", ResourceType.WAV)

    # Writing (clean for media players):
    write_wav(wav, "output.wav", ResourceType.WAV_DEOB)

    # Get playable bytes (for Qt media player, etc.):
    playable_bytes = get_playable_bytes(wav)

References:
----------
        See wav_data module docstring for engine addresses (K1 + TSL TODO).
        Derivations and Other Implementations:
        ----------
        https://github.com/th3w1zard1/KotOR.js/tree/master/src/audio/AudioFile.ts:164-205


"""

from __future__ import annotations

from typing import TYPE_CHECKING

from pykotor.resource.formats.wav.io_wav import WAVBinaryReader, WAVBinaryWriter
from pykotor.resource.formats.wav.io_wav_standard import WAVStandardWriter
from pykotor.resource.formats.wav.wav_data import AudioFormat
from pykotor.resource.type import ResourceType

if TYPE_CHECKING:
    from pykotor.resource.formats.wav.wav_data import WAV
    from pykotor.resource.type import SOURCE_TYPES, TARGET_TYPES


def read_wav(
    source: SOURCE_TYPES,
    offset: int = 0,
    size: int | None = None,
) -> WAV:
    """Returns a WAV instance from the source with automatic deobfuscation.

    The function automatically detects and removes obfuscation headers if present.
    The returned WAV object contains clean, deobfuscated data.

    Supported formats:
        - Standard RIFF/WAVE (streamvoice, dialog audio)
        - SFX with 470-byte header (streammusic, sound effects)
        - MP3-in-WAV (some music files with RIFF size == 50)

    Args:
        source: The source of the data (file path, bytes, or file-like object)
        offset: The byte offset of the file inside the data
        size: Number of bytes to read. If not specified, uses the whole stream.

    Raises:
        FileNotFoundError: If the file could not be found.
        IsADirectoryError: If the specified path is a directory.
        PermissionError: If the file could not be accessed.
        ValueError: If the file was corrupted or format unrecognized.

    Returns:
        A WAV instance (automatically deobfuscated).
    """
    return WAVBinaryReader(source, offset, size or 0).load()


def write_wav(
    wav: WAV,
    target: TARGET_TYPES,
    file_format: ResourceType = ResourceType.WAV,
):
    """Writes the WAV data to the target location.

    If file_format == ResourceType.WAV, the data will be obfuscated based on
    the WAV's type (SFX adds header, VO is unchanged) for game compatibility.

    If file_format == ResourceType.WAV_DEOB, writes clean RIFF/WAVE format
    playable by standard media players.

    Args:
        wav: The WAV file being written.
        target: The location to write the data to.
        file_format: WAV for game format, WAV_DEOB for clean playable format.

    Raises:
        IsADirectoryError: If the specified path is a directory.
        PermissionError: If the file could not be written.
    """
    if file_format == ResourceType.WAV:
        WAVBinaryWriter(wav, target).write()
    else:
        WAVStandardWriter(wav, target).write()


def bytes_wav(
    wav: WAV,
    file_format: ResourceType = ResourceType.WAV,
) -> bytes:
    """Returns the WAV data as a bytes object.

    If file_format == ResourceType.WAV, returns obfuscated format for game use.
    If file_format == ResourceType.WAV_DEOB, returns clean playable format.

    Args:
        wav: The target WAV object.
        file_format: WAV for game format, WAV_DEOB for clean playable format.

    Returns:
        The WAV data as bytes.
    """
    data = bytearray()
    write_wav(wav, data, file_format)
    return bytes(data)


def get_playable_bytes(wav: WAV) -> bytes:
    """Returns playable audio bytes for media player use.

    This is the preferred method for getting audio data that can be played
    by Qt's QMediaPlayer or other standard audio players.

    For MP3 format: Returns raw MP3 bytes
    For WAVE format: Returns clean RIFF/WAVE structure

    Args:
        wav: The WAV object to convert

    Returns:
        Audio bytes playable by standard media players

    References:
        https://github.com/th3w1zard1/KotOR.js/tree/master/src/audio/AudioFile.ts:164-205 - getPlayableByteStream()
    """
    return bytes_wav(wav, ResourceType.WAV_DEOB)


def detect_audio_type(wav: WAV) -> str:
    """Returns the file extension appropriate for this audio's actual format.

    Useful for saving to temp files with correct extension.

    Args:
        wav: The WAV object to check

    Returns:
        "mp3" for MP3 format, "wav" for WAVE format

    References:
        https://github.com/th3w1zard1/KotOR.js/tree/master/src/audio/AudioFile.ts:348-354 - getExportExtension()
    """
    if wav.audio_format == AudioFormat.MP3:
        return "mp3"
    return "wav"



"""Binary NCS (NWScript compiled script) read/write and VM validation."""

from __future__ import annotations

from typing import TYPE_CHECKING

from pykotor.resource.formats.ncs.ncs_data import (
    NCS,
    NCSByteCode,
    NCSInstruction,
    NCSInstructionType,
    NCSInstructionTypeValue,
)
from pykotor.resource.formats.ncs.vm_validation import validate_ncs_for_vm
from pykotor.resource.type import ResourceReader, ResourceWriter, autoclose

if TYPE_CHECKING:
    from pykotor.resource.type import SOURCE_TYPES, TARGET_TYPES

# NCS file format constants
NCS_HEADER_MAGIC_BYTE = 0x42
NCS_HEADER_SIZE = 13  # "NCS " (4) + "V1.0" (4) + magic_byte (1) + size (4)


class NCSBinaryReader(ResourceReader):
    """Reads NCS (NWScript Compiled Script) files.

    NCS files contain compiled bytecode for NWScript, the scripting language used in KotOR.
    Instructions include operations, constants, function calls, jumps, and control flow.

    References:
    ----------
        See ncs_data module docstring for full engine addresses (K1 + TSL TODO).
        CResNCS::CResNCS (K1: 0x005d4c30), destructors (0x005d4c50, 0x005d4c90), ReadScriptFile (0x005d2260),
        ReadScriptsFromGff (0x004ebf20), ExecuteCommandExecuteScript (0x00535b70). NCS format: "NCS " type, "V1.0", magic 0x42.
        Note: Engine loads NCS as resources and executes via NWScript VM.


    """

    def __init__(
        self,
        source: SOURCE_TYPES,
        offset: int = 0,
        size: int = 0,
        *,
        validate_for_vm: bool = False,
    ):
        super().__init__(source, offset, size)
        self._ncs: NCS | None = None
        self._instructions: dict[int, NCSInstruction] = {}
        self._jumps: list[tuple[NCSInstruction, int]] = []
        self._validate_for_vm: bool = validate_for_vm

    @autoclose
    def load(self, *, auto_close: bool = True) -> NCS:
        """Loads an NCS file from the reader.

        Returns:
        -------
            NCS: The loaded NCS object

        Raises:
            ValueError - Corrupt NCS.
            OSError - some operating system issue occurred.

        Processing Logic:
        ----------------
            - Reads the file type and version headers
            - Reads each instruction from the file into a dictionary
            - Resolves jump offsets to reference the target instructions
            - Adds the instructions to the NCS object
            - Optionally closes the reader.
        """
        self._ncs = NCS()

        file_type = self._reader.read_string(4)
        file_version = self._reader.read_string(4)

        if file_type != "NCS ":
            msg = "The file type that was loaded is invalid."
            raise ValueError(msg)

        if file_version != "V1.0":
            msg = "The NCS version that was loaded is not supported."
            raise ValueError(msg)

        self._instructions = {}  # offset -> instruction
        self._jumps = []

        # Read the header fields

        magic_byte = self._reader.read_uint8()  # Position 8
        total_size = self._reader.read_uint32(big=True)  # Positions 9-12: Total file size

        # Validate header
        if magic_byte != NCS_HEADER_MAGIC_BYTE:
            msg = f"Invalid NCS header magic byte: expected 0x{NCS_HEADER_MAGIC_BYTE:02X}, got 0x{magic_byte:02X}"
            raise ValueError(msg)

        # Validate size field
        actual_file_size = self._reader.size()
        if total_size > actual_file_size:
            msg = f"NCS size field ({total_size}) is larger than actual file size ({actual_file_size}). File may be corrupted or truncated."
            raise ValueError(msg)

        # Check for empty or minimal NCS files
        if total_size <= NCS_HEADER_SIZE:
            # File has only a header, no instructions
            # This is technically valid but unusual
            self._ncs.instructions = []
            return self._ncs

        # Now at position 13, read instructions until we reach total_size
        # total_size is the end position (includes the header)
        code_end_position = total_size

        # Safety: don't read beyond actual file size
        safe_end_position = min(code_end_position, actual_file_size)

        while self._reader.position() < safe_end_position and self._reader.remaining() > 0:
            offset = self._reader.position()

            try:
                self._instructions[offset] = self._read_instruction()
            except ValueError as e:
                error_msg = str(e)

                # Check if this is zero-padding that slipped through due to incorrect size field
                if "Unknown NCS bytecode 0x00" in error_msg:
                    # Peek ahead to confirm this is just padding
                    self._reader.seek(offset)

                    # Read remaining bytes up to the safe end position
                    bytes_to_check = min(self._reader.remaining(), safe_end_position - offset)
                    remaining_data = self._reader.read_bytes(bytes_to_check)

                    if all(b == 0 for b in remaining_data):
                        # This is zero-padding - the size field incorrectly includes padding
                        import sys  # noqa: PLC0415

                        print(
                            f"Warning: NCS file has incorrect size field (includes zero-padding). "
                            f"Size field: {total_size}, actual code ends at: {offset}, "
                            f"found {len(remaining_data)} bytes of padding",
                            file=sys.stderr,
                        )
                        break

                    # Not all zeros - this is genuinely corrupted data
                    # Show diagnostic information
                    self._reader.seek(offset)
                    diagnostic_bytes = self._reader.read_bytes(min(32, self._reader.remaining()))
                    diagnostic_hex = " ".join(f"{b:02X}" for b in diagnostic_bytes)

                    enhanced_msg = (
                        f"{error_msg}\n"
                        f"  File size field: {total_size}, current offset: {offset}\n"
                        f"  Next 32 bytes (hex): {diagnostic_hex}\n"
                        f"  This indicates the NCS file is genuinely corrupted or uses an unknown format variant."
                    )
                    raise ValueError(enhanced_msg) from e

                # Re-raise other errors with additional context
                enhanced_msg = f"Failed to parse NCS instruction at offset {offset}: {error_msg}"
                raise ValueError(enhanced_msg) from e

        for instruction, jumpToOffset in self._jumps:
            instruction.jump = self._instructions[jumpToOffset]

        self._ncs.instructions = list(self._instructions.values())

        # Optional strict validation for VM compatibility (can reject valid real-world scripts).
        if self._validate_for_vm:
            validate_ncs_for_vm(self._ncs)

        return self._ncs

    def _read_instruction(self) -> NCSInstruction:
        """Reads an instruction from the bytecode reader.

        Args:
        ----
            self: {The class instance}: Provides access to the bytecode reader

        Returns:
        -------
            instruction: {An NCSInstruction object}: The instruction read from the bytecode

        Processing Logic:
        ----------------
            - Reads the byte code and qualifier from the reader
            - Determines the instruction type from these values
            - Initializes an NCSInstruction object
            - Reads arguments from the reader based on the instruction type
            - Handles jump offsets
            - Returns the completed instruction
        """
        instruction_offset = self._reader.position()
        byte_code_value = self._reader.read_uint8()
        qualifier = self._reader.read_uint8()

        # Try to convert to NCSByteCode enum
        try:
            byte_code = NCSByteCode(byte_code_value)
        except ValueError as e:
            # Provide detailed diagnostic information for unknown bytecodes
            # Read some context bytes around the error position
            context_size = 16
            context_start = max(0, instruction_offset - context_size)
            current_pos = self._reader.position()

            self._reader.seek(context_start)
            context_bytes = self._reader.read_bytes(min(context_size * 2, self._reader.size() - context_start))
            context_hex = " ".join(f"{b:02X}" for b in context_bytes)

            self._reader.seek(current_pos)  # Restore position

            msg = (
                f"Unknown NCS bytecode 0x{byte_code_value:02X} with qualifier 0x{qualifier:02X} at offset {instruction_offset}.\n"
                f"  Context (hex): {context_hex}\n"
                f"  This NCS file may be corrupted, from an unsupported NCS variant, or have an incorrect size field."
            )
            raise ValueError(msg) from e

        type_value = NCSInstructionTypeValue(byte_code, qualifier)

        instruction = NCSInstruction()
        instruction.offset = instruction_offset

        # Special handling for RESERVED opcode - it appears with various qualifiers
        # Treat all RESERVED variants as simple 2-byte no-ops
        if byte_code == NCSByteCode.RESERVED:
            # Use RESERVED (0x00, 0x00) as the canonical type regardless of actual qualifier
            instruction.ins_type = NCSInstructionType.RESERVED
        else:
            try:
                instruction.ins_type = NCSInstructionType(type_value)
            except ValueError as e:
                # Unknown bytecode/qualifier combination - the bytecode exists but this specific
                # combination with the qualifier is not recognized
                msg = (
                    f"Unknown NCS instruction type combination: "
                    f"bytecode=0x{byte_code_value:02X} ({byte_code.name}), "
                    f"qualifier=0x{qualifier:02X} at offset {instruction_offset}.\n"
                    f"  The bytecode is recognized but this qualifier combination is not supported."
                )
                raise ValueError(msg) from e

        if instruction.ins_type in {
            NCSInstructionType.CPDOWNSP,
            NCSInstructionType.CPTOPSP,
            NCSInstructionType.CPDOWNBP,
            NCSInstructionType.CPTOPBP,
        }:
            instruction.args.extend([self._reader.read_int32(big=True), self._reader.read_uint16(big=True)])

        elif instruction.ins_type == NCSInstructionType.CONSTI:
            instruction.args.extend([self._reader.read_uint32(big=True)])

        elif instruction.ins_type == NCSInstructionType.CONSTF:
            instruction.args.extend([self._reader.read_single(big=True)])

        elif instruction.ins_type == NCSInstructionType.CONSTS:
            length = self._reader.read_uint16(big=True)
            instruction.args.extend([self._reader.read_string(length)])

        elif instruction.ins_type == NCSInstructionType.CONSTO:
            # Object constants are stored as signed 32-bit integers, not 16-bit
            # See DeNCS Decoder.java case 4, subcase 6 (OBJECT type uses readSignedInt)
            instruction.args.extend([self._reader.read_int32(big=True)])

        elif instruction.ins_type == NCSInstructionType.ACTION:
            instruction.args.extend([self._reader.read_uint16(big=True), self._reader.read_uint8(big=True)])

        elif instruction.ins_type == NCSInstructionType.MOVSP:
            instruction.args.extend([self._reader.read_int32(big=True)])

        elif instruction.ins_type in {
            NCSInstructionType.JMP,
            NCSInstructionType.JSR,
            NCSInstructionType.JZ,
            NCSInstructionType.JNZ,
        }:
            jumpOffset = self._reader.read_int32(big=True) + self._reader.position() - 6
            self._jumps.append((instruction, jumpOffset))

        elif instruction.ins_type == NCSInstructionType.DESTRUCT:
            instruction.args.extend(
                [
                    self._reader.read_uint16(big=True),
                    self._reader.read_int16(big=True),
                    self._reader.read_uint16(big=True),
                ],
            )

        elif instruction.ins_type in {
            NCSInstructionType.DECxSP,
            NCSInstructionType.INCxSP,
            NCSInstructionType.DECxBP,
            NCSInstructionType.INCxBP,
        }:
            instruction.args.extend([self._reader.read_uint32(big=True)])

        elif instruction.ins_type == NCSInstructionType.STORE_STATE:
            instruction.args.extend([self._reader.read_uint32(big=True), self._reader.read_uint32(big=True)])

        elif instruction.ins_type in {
            NCSInstructionType.EQUALTT,
            NCSInstructionType.NEQUALTT,
        }:
            # Struct equality comparisons include a size field
            # See DeNCS Decoder.java case 11/12 with qualifier 0x24 (36 = StructStruct)
            instruction.args.extend([self._reader.read_uint16(big=True)])

        elif instruction.ins_type in {  # noqa: SIM114
            NCSInstructionType.LOGANDII,
            NCSInstructionType.LOGORII,
        }:
            ...

        elif instruction.ins_type in {  # noqa: SIM114
            NCSInstructionType.INCORII,
            NCSInstructionType.EXCORII,
        }:
            ...

        elif instruction.ins_type == NCSInstructionType.BOOLANDII:  # noqa: SIM114
            ...

        elif instruction.ins_type in {  # noqa: SIM114
            NCSInstructionType.EQUALII,
            NCSInstructionType.EQUALFF,
            NCSInstructionType.EQUALOO,
            NCSInstructionType.EQUALEFFEFF,
            NCSInstructionType.EQUALEVTEVT,
            NCSInstructionType.EQUALLOCLOC,
            NCSInstructionType.EQUALTALTAL,
            NCSInstructionType.EQUALSS,
        }:
            ...

        elif instruction.ins_type in {  # noqa: SIM114
            NCSInstructionType.NEQUALII,
            NCSInstructionType.NEQUALFF,
            NCSInstructionType.NEQUALOO,
            NCSInstructionType.NEQUALEFFEFF,
            NCSInstructionType.NEQUALEVTEVT,
            NCSInstructionType.NEQUALLOCLOC,
            NCSInstructionType.NEQUALTALTAL,
            NCSInstructionType.NEQUALSS,
        }:
            ...

        elif instruction.ins_type in {  # noqa: SIM114
            # NCSInstructionType.GEQxx,
            NCSInstructionType.GEQII,
            NCSInstructionType.GEQFF,
            # NCSInstructionType.GTxx,
            NCSInstructionType.GTII,
            NCSInstructionType.GTFF,
        }:
            ...

        elif instruction.ins_type in {  # noqa: SIM114
            # NCSInstructionType.LTxx,
            NCSInstructionType.LTII,
            NCSInstructionType.LTFF,
            # NCSInstructionType.LExx,
            NCSInstructionType.LEQII,
            NCSInstructionType.LEQFF,
        }:
            ...

        elif instruction.ins_type in {  # noqa: SIM114
            NCSInstructionType.SHLEFTII,
            NCSInstructionType.SHRIGHTII,
            NCSInstructionType.USHRIGHTII,
        }:
            ...

        elif instruction.ins_type in {  # noqa: SIM114
            NCSInstructionType.ADDII,
            NCSInstructionType.ADDFF,
            NCSInstructionType.ADDFI,
            NCSInstructionType.ADDIF,
            NCSInstructionType.ADDSS,
            NCSInstructionType.ADDVV,
        }:
            ...

        elif instruction.ins_type in {  # noqa: SIM114
            NCSInstructionType.SUBII,
            NCSInstructionType.SUBFF,
            NCSInstructionType.SUBFI,
            NCSInstructionType.SUBIF,
            NCSInstructionType.SUBVV,
        }:
            ...

        elif instruction.ins_type in {  # noqa: SIM114
            NCSInstructionType.MULII,
            NCSInstructionType.MULFF,
            NCSInstructionType.MULFI,
            NCSInstructionType.MULIF,
            NCSInstructionType.MULFV,
            NCSInstructionType.MULVF,
        }:
            ...

        elif instruction.ins_type in {  # noqa: SIM114
            NCSInstructionType.DIVII,
            NCSInstructionType.DIVFF,
            NCSInstructionType.DIVFI,
            NCSInstructionType.DIVIF,
            NCSInstructionType.DIVFV,
            NCSInstructionType.DIVVF,
        }:
            ...

        elif instruction.ins_type == NCSInstructionType.MODII:  # noqa: SIM114
            ...

        elif instruction.ins_type in {  # noqa: SIM114
            # NCSInstructionType.NEGx,
            NCSInstructionType.NEGI,
            NCSInstructionType.NEGF,
        }:
            ...

        elif (
            instruction.ins_type == NCSInstructionType.COMPI
            or instruction.ins_type
            in {  # noqa: SIM114
                # NCSInstructionType.STORE_STATEALL,
            }
        ):
            ...

        elif instruction.ins_type == NCSInstructionType.RETN:  # noqa: SIM114
            ...

        elif instruction.ins_type == NCSInstructionType.NOTI:  # noqa: SIM114
            ...

        elif instruction.ins_type in {  # noqa: SIM114
            NCSInstructionType.SAVEBP,
            NCSInstructionType.RESTOREBP,
        }:
            ...

        elif instruction.ins_type == NCSInstructionType.NOP:  # noqa: SIM114
            ...

        elif instruction.ins_type in {  # noqa: SIM114
            NCSInstructionType.RESERVED,
            NCSInstructionType.RESERVED_01,
        }:
            # Reserved/unknown opcodes - treat as 2-byte no-ops with no arguments
            ...

        elif instruction.ins_type in {
            NCSInstructionType.RSADDI,
            NCSInstructionType.RSADDF,
            NCSInstructionType.RSADDO,
            NCSInstructionType.RSADDS,
            NCSInstructionType.RSADDEFF,  # ???
            NCSInstructionType.RSADDEVT,  # ???
            NCSInstructionType.RSADDLOC,  # ???
            NCSInstructionType.RSADDTAL,  # ???
        }:
            ...

        else:
            msg = f"Tried to read unsupported instruction '{instruction.ins_type.name}' to NCS"
            raise ValueError(msg)

        return instruction


class NCSBinaryWriter(ResourceWriter):
    def __init__(
        self,
        ncs: NCS,
        target: TARGET_TYPES,
    ):
        super().__init__(target)
        self._ncs: NCS = ncs
        self._offsets: dict[int, int] = {}
        self._sizes: dict[int, int] = {}

    @autoclose
    def write(self, *, auto_close: bool = True):
        """Writes the NCS file.

        Args:
        ----
            auto_close (bool): Whether to automatically close the writer.

        Processing Logic:
        ----------------
            - Calculates offset and size for each instruction
            - Writes header with file type and total size
            - Writes each instruction using pre-calculated offset and size
            - Closes writer if auto_close is True.
        """
        offset = NCS_HEADER_SIZE
        for instruction in self._ncs.instructions:
            inst_id = id(instruction)
            instruction_size = self.determine_size(instruction)
            self._sizes[inst_id] = instruction_size
            self._offsets[inst_id] = offset
            offset += instruction_size

        self._writer.write_string("NCS ")
        self._writer.write_string("V1.0")
        self._writer.write_uint8(NCS_HEADER_MAGIC_BYTE)
        self._writer.write_uint32(offset, big=True)

        for instruction in self._ncs.instructions:
            self._write_instruction(instruction)

    def determine_size(self, instruction: NCSInstruction) -> int:  # TODO(th3w1zard1): This function is unfinished and is missing defs.
        """Determines the size of an NCS instruction in bytes.

        Based on DeNCS Decoder.java readCommand method which shows complete instruction formats.

        Args:
        ----
            instruction: NCSInstruction - The instruction to determine size for

        Returns:
        -------
            int - The size of the instruction in bytes
        """
        size = 2  # Base size for opcode and qualifier bytes

        # Copy operations: 1 byte qualifier + 4 bytes offset + 2 bytes size = 7 bytes additional
        # DeNCS case 1, 3, 38, 39
        if instruction.ins_type in {
            NCSInstructionType.CPDOWNSP,
            NCSInstructionType.CPTOPSP,
            NCSInstructionType.CPDOWNBP,
            NCSInstructionType.CPTOPBP,
        }:
            size += 6  # 1 + 4 + 2 - 1 (qualifier already in base) = 6

        # STORE_STATE: 1 byte qualifier + 4 bytes + 4 bytes = 9 bytes additional
        # DeNCS case 44
        elif instruction.ins_type == NCSInstructionType.STORE_STATE:
            size += 8  # 1 + 4 + 4 - 1 = 8

        # Struct equality: 1 byte qualifier + 2 bytes size = 3 bytes additional
        # DeNCS case 11, 12 with qualifier 0x24
        elif instruction.ins_type in {
            NCSInstructionType.NEQUALTT,
            NCSInstructionType.EQUALTT,
        }:
            size += 2  # 1 + 2 - 1 = 2

        # Jump/Move operations: 1 byte qualifier + 4 bytes offset = 5 bytes additional
        # DeNCS case 27, 29, 30, 31, 35, 36, 37, 40, 41
        elif instruction.ins_type in {
            NCSInstructionType.MOVSP,
            NCSInstructionType.JMP,
            NCSInstructionType.JSR,
            NCSInstructionType.JZ,
            NCSInstructionType.JNZ,
            NCSInstructionType.DECxSP,
            NCSInstructionType.INCxSP,
            NCSInstructionType.DECxBP,
            NCSInstructionType.INCxBP,
        }:
            size += 4  # 1 + 4 - 1 = 4

        # Constants with 4-byte values
        # DeNCS case 4 with subcases for int/float/object
        elif instruction.ins_type in {
            NCSInstructionType.CONSTI,
            NCSInstructionType.CONSTF,
            NCSInstructionType.CONSTO,
        }:
            size += 4  # 4 bytes for the constant value

        # String constant: 2 bytes length + string data
        # DeNCS case 4 subcase 5 (string)
        elif instruction.ins_type == NCSInstructionType.CONSTS:
            size += 2 + len(instruction.args[0])  # 2 bytes length prefix + string

        # ACTION: 2 bytes routine number + 1 byte argument count
        # DeNCS case 5
        elif instruction.ins_type == NCSInstructionType.ACTION:
            size += 3  # 2 + 1 = 3

        # DESTRUCT: 1 byte qualifier + 2 bytes + 2 bytes + 2 bytes = 7 bytes additional
        # DeNCS case 33
        elif instruction.ins_type == NCSInstructionType.DESTRUCT:
            size += 6  # 1 + 2 + 2 + 2 - 1 = 6

        # All other instructions have just opcode + qualifier (2 bytes total)
        # This includes: RSADD variants, logical/arithmetic ops, RETN, SAVEBP, RESTOREBP, etc.
        # DeNCS case 2, 6-10, 13-26, 32, 34, 42, 43, 45

        return size

    def _write_instruction(self, instruction: NCSInstruction):  # TODO(th3w1zard1): This function is unfinished and is missing defs.
        """Writes an instruction to the NCS binary stream.

        Args:
        ----
            instruction (NCSInstruction): The instruction to write

        Processing Logic:
        ----------------
            - Writes instruction type and qualifier bytes
            - Writes instruction arguments based on type
                - Integer, float, string, object ID
                - Relative jump offsets
            - Raises error for unsupported instructions
        """

        def to_signed_32bit(n: int) -> int:  # FIXME(th3w1zard1): Presumably this issue happens further up the call stack, fix later.
            """Convert unsigned 32-bit integer representation to signed.

            Handles edge cases where values may be stored as unsigned but need
            to be interpreted/written as signed two's complement integers.
            Python's int type can represent any value, but when reading from
            binary data or external sources, values may need conversion.

            Args:
            ----
                n: Integer value, possibly in unsigned representation

            Returns:
            -------
                Signed 32-bit integer value
            """
            # Convert to signed if value is in upper half of unsigned 32-bit range
            if n >= 2**31:
                n -= 2**32
            return n

        def to_signed_16bit(n: int) -> int:  # FIXME(th3w1zard1): Only seen this issue happen with 32bit but better safe than sorry, remove this once above issue is fixed.
            """Convert unsigned 16-bit integer representation to signed.

            Similar to to_signed_32bit but for 16-bit values.

            Args:
            ----
                n: Integer value, possibly in unsigned representation

            Returns:
            -------
                Signed 16-bit integer value
            """
            if n >= 2**15:
                n -= 2**16
            return n

        self._writer.write_uint8(int(instruction.ins_type.value.byte_code))
        self._writer.write_uint8(int(instruction.ins_type.value.qualifier))

        # Handle instruction-specific arguments
        if instruction.ins_type in {
            NCSInstructionType.DECxSP,
            NCSInstructionType.INCxSP,
            NCSInstructionType.DECxBP,
            NCSInstructionType.INCxBP,
        }:
            self._writer.write_int32(to_signed_32bit(instruction.args[0]), big=True)

        elif instruction.ins_type in {NCSInstructionType.CPDOWNSP, NCSInstructionType.CPTOPSP, NCSInstructionType.CPDOWNBP, NCSInstructionType.CPTOPBP}:
            self._writer.write_int32(instruction.args[0], big=True)
            # Size argument: typically 4 for most types, 12 for vectors (3 floats)
            # The args[1] should contain the actual size value from compilation
            size_value = instruction.args[1] if len(instruction.args) > 1 else 4
            self._writer.write_uint16(size_value, big=True)

        elif instruction.ins_type == NCSInstructionType.CONSTF:
            self._writer.write_single(instruction.args[0], big=True)
        elif instruction.ins_type == NCSInstructionType.CONSTO:
            # Object constants are stored as signed 32-bit integers
            # See DeNCS Decoder.java case 4, subcase 6
            self._writer.write_int32(instruction.args[0], big=True)
        elif instruction.ins_type == NCSInstructionType.CONSTI:
            # Integer constants stored as unsigned 32-bit per DeNCS Decoder.java line 137
            # Convert to signed representation for struct packing
            self._writer.write_int32(to_signed_32bit(instruction.args[0]), big=True)
        elif instruction.ins_type == NCSInstructionType.CONSTS:
            # CONSTS with string length and string data
            self._writer.write_string(instruction.args[0], big=True, prefix_length=2)

        elif instruction.ins_type == NCSInstructionType.ACTION:
            # ACTION with routine number and argument count
            self._writer.write_uint16(instruction.args[0], big=True)
            self._writer.write_uint8(instruction.args[1], big=True)

        elif instruction.ins_type in {  # noqa: SIM114
            NCSInstructionType.LOGANDII,
            NCSInstructionType.LOGORII,
        }:
            ...

        elif instruction.ins_type in {  # noqa: SIM114
            NCSInstructionType.INCORII,
            NCSInstructionType.EXCORII,
        }:
            ...

        elif instruction.ins_type == NCSInstructionType.BOOLANDII:  # noqa: SIM114
            ...

        elif instruction.ins_type in {  # noqa: SIM114
            NCSInstructionType.EQUALII,
            NCSInstructionType.EQUALFF,
            NCSInstructionType.EQUALOO,
            NCSInstructionType.EQUALEFFEFF,
            NCSInstructionType.EQUALEVTEVT,
            NCSInstructionType.EQUALLOCLOC,
            NCSInstructionType.EQUALTALTAL,
            NCSInstructionType.EQUALSS,
        }:
            ...

        elif instruction.ins_type in {
            NCSInstructionType.NEQUALII,
            NCSInstructionType.NEQUALFF,
            NCSInstructionType.NEQUALOO,
            NCSInstructionType.NEQUALEFFEFF,
            NCSInstructionType.NEQUALEVTEVT,
            NCSInstructionType.NEQUALLOCLOC,
            NCSInstructionType.NEQUALTALTAL,
            NCSInstructionType.NEQUALSS,
        }:
            ...

        elif instruction.ins_type in {
            NCSInstructionType.EQUALTT,
            NCSInstructionType.NEQUALTT,
        }:
            self._writer.write_uint16(instruction.args[0], big=True)

        elif instruction.ins_type in {  # noqa: SIM114
            # NCSInstructionType.GEQxx,
            NCSInstructionType.GEQII,
            NCSInstructionType.GEQFF,
            # NCSInstructionType.GTxx,
            NCSInstructionType.GTII,
            NCSInstructionType.GTFF,
        }:
            ...

        elif instruction.ins_type in {  # noqa: SIM114
            # NCSInstructionType.LTxx,
            NCSInstructionType.LTII,
            NCSInstructionType.LTFF,
            # NCSInstructionType.LExx,
            NCSInstructionType.LEQII,
            NCSInstructionType.LEQFF,
        }:
            ...

        elif instruction.ins_type in {  # noqa: SIM114
            NCSInstructionType.SHLEFTII,
            NCSInstructionType.SHRIGHTII,
            NCSInstructionType.USHRIGHTII,
        }:
            ...

        elif instruction.ins_type in {  # noqa: SIM114
            NCSInstructionType.ADDII,
            NCSInstructionType.ADDFF,
            NCSInstructionType.ADDFI,
            NCSInstructionType.ADDIF,
            NCSInstructionType.ADDSS,
            NCSInstructionType.ADDVV,
        }:
            ...

        elif instruction.ins_type in {  # noqa: SIM114
            NCSInstructionType.SUBII,
            NCSInstructionType.SUBFF,
            NCSInstructionType.SUBFI,
            NCSInstructionType.SUBIF,
            NCSInstructionType.SUBVV,
        }:
            ...

        elif instruction.ins_type in {  # noqa: SIM114
            NCSInstructionType.MULII,
            NCSInstructionType.MULFF,
            NCSInstructionType.MULFI,
            NCSInstructionType.MULIF,
            NCSInstructionType.MULFV,
            NCSInstructionType.MULVF,
        }:
            ...

        elif instruction.ins_type in {  # noqa: SIM114
            NCSInstructionType.DIVII,
            NCSInstructionType.DIVFF,
            NCSInstructionType.DIVFI,
            NCSInstructionType.DIVIF,
            NCSInstructionType.DIVFV,
            NCSInstructionType.DIVVF,
        }:
            ...

        elif instruction.ins_type == NCSInstructionType.MODII:  # noqa: SIM114
            ...

        elif instruction.ins_type in {  # noqa: SIM114
            # NCSInstructionType.NEGx,
            NCSInstructionType.NEGI,
            NCSInstructionType.NEGF,
        }:
            ...

        elif instruction.ins_type == NCSInstructionType.COMPI:
            ...

        elif instruction.ins_type == NCSInstructionType.MOVSP:
            # MOVSP to adjust the stack pointer
            self._writer.write_int32(to_signed_32bit(instruction.args[0]), big=True)

        elif (
            instruction.ins_type
            in {
                # NCSInstructionType.STORE_STATEALL,
            }
        ):
            ...

        elif instruction.ins_type in {
            NCSInstructionType.JMP,
            NCSInstructionType.JSR,
            NCSInstructionType.JZ,
            NCSInstructionType.JNZ,
        }:
            jump = instruction.jump
            assert jump is not None, f"{instruction} has a NoneType jump."
            instruction_id = id(instruction)
            jump_id = id(jump)
            relative = self._offsets[jump_id] - self._offsets[instruction_id]
            self._writer.write_int32(to_signed_32bit(relative), big=True)

        elif instruction.ins_type == NCSInstructionType.RETN:
            ...

        elif instruction.ins_type == NCSInstructionType.DESTRUCT:
            self._writer.write_uint16(instruction.args[0], big=True)
            self._writer.write_int16(to_signed_16bit(instruction.args[1]), big=True)
            self._writer.write_uint16(instruction.args[2], big=True)

        elif instruction.ins_type == NCSInstructionType.NOTI:  # noqa: SIM114
            ...

        elif instruction.ins_type in {NCSInstructionType.SAVEBP, NCSInstructionType.RESTOREBP}:
            ...

        elif instruction.ins_type == NCSInstructionType.STORE_STATE:
            self._writer.write_uint32(instruction.args[0], big=True)
            self._writer.write_uint32(instruction.args[1], big=True)

        elif instruction.ins_type == NCSInstructionType.NOP:  # noqa: SIM114
            ...

        elif instruction.ins_type in {  # noqa: SIM114
            NCSInstructionType.RESERVED,
            NCSInstructionType.RESERVED_01,
        }:
            # Reserved/unknown opcodes - no arguments to write
            ...

        elif instruction.ins_type in {
            NCSInstructionType.RSADDI,
            NCSInstructionType.RSADDF,
            NCSInstructionType.RSADDO,
            NCSInstructionType.RSADDS,
            NCSInstructionType.RSADDEFF,  # ???
            NCSInstructionType.RSADDEVT,  # ???
            NCSInstructionType.RSADDLOC,  # ???
            NCSInstructionType.RSADDTAL,  # ???
        }:
            ...

        else:
            msg = f"Tried to write unsupported instruction ({instruction.ins_type.name}) to NCS"
            raise ValueError(msg)



"""High-level reference extraction and referrer search using reference_config.

Provides extract_references (from a single resource) and find_referrers (across
an installation) for scripts, conversations, tags, and template resrefs.
"""

from __future__ import annotations

from dataclasses import dataclass
from typing import TYPE_CHECKING, Any

from pykotor.resource.formats.gff.gff_data import GFFFieldType, GFFList, GFFStruct
from pykotor.tools.reference_config import (
    DLG_SCRIPT_PATHS,
    get_conversation_fields_for_type,
    get_script_fields_for_type,
    get_tag_fields_for_type,
    get_template_resref_fields_for_type,
)
from pykotor.tools.reference_finder import (
    ReferenceSearchResult,
    find_conversation_references,
    find_resref_references,
    find_script_references,
    find_tag_references,
)

if TYPE_CHECKING:
    from pykotor.extract.installation import Installation
    from pykotor.resource.formats.gff import GFF


@dataclass(frozen=True)
class ExtractedReference:
    """A single reference extracted from a GFF (script, conversation, tag, template_resref)."""

    ref_kind: str
    value: str
    field_path: str


def _value_to_str(value: Any) -> str:
    """Normalize GFF field value to string for ResRef/String."""
    if value is None or value == "":
        return ""
    if hasattr(value, "get") and callable(getattr(value, "get", None)):
        return str(value.get()) if value.get() else ""
    return str(value).strip()


def _collect_from_struct(
    struct: GFFStruct,
    path_prefix: str,
    file_type: str,
    script_fields: set[str],
    tag_fields: set[str],
    template_fields: set[str],
    conversation_fields: set[str],
    out: list[ExtractedReference],
) -> None:
    """Walk a GFF struct and append ExtractedReference for configured fields."""
    for label, field_type, value in struct:
        if value is None:
            continue
        field_path = f"{path_prefix}.{label}" if path_prefix else label

        if field_type == GFFFieldType.ResRef:
            s = _value_to_str(value)
            if s:
                if label in script_fields:
                    out.append(ExtractedReference("script", s, field_path))
                if label in template_fields:
                    out.append(ExtractedReference("template_resref", s, field_path))
                if label in conversation_fields:
                    out.append(ExtractedReference("conversation", s, field_path))
        elif field_type == GFFFieldType.String:
            s = _value_to_str(value)
            if s and label in tag_fields:
                out.append(ExtractedReference("tag", s, field_path))
        elif field_type == GFFFieldType.Struct and isinstance(value, GFFStruct):
            _collect_from_struct(
                value,
                field_path,
                file_type,
                script_fields,
                tag_fields,
                template_fields,
                conversation_fields,
                out,
            )
        elif field_type == GFFFieldType.List and isinstance(value, GFFList):
            for idx, item in enumerate(value):
                if isinstance(item, GFFStruct):
                    list_path = f"{field_path}[{idx}]"
                    _collect_from_struct(
                        item,
                        list_path,
                        file_type,
                        script_fields,
                        tag_fields,
                        template_fields,
                        conversation_fields,
                        out,
                    )


def extract_references(
    gff: GFF,
    file_type: str,
) -> list[ExtractedReference]:
    """Extract all script, conversation, tag, and template_resref references from a GFF.

    Uses pykotor.tools.reference_config for field maps. DLG nested lists (StartingList,
    EntryList, ReplyList) are walked so script/condition refs inside entries/replies
    are included.

    Args:
    ----
        gff: Loaded GFF (e.g. from read_gff(resource.data()))
        file_type: Uppercase extension (e.g. "UTC", "DLG", "ARE")

    Returns:
    -------
        List of ExtractedReference (ref_kind, value, field_path)
    """
    script_fields = get_script_fields_for_type(file_type)
    tag_fields = get_tag_fields_for_type(file_type)
    template_fields = get_template_resref_fields_for_type(file_type)
    conversation_fields = get_conversation_fields_for_type(file_type)

    out: list[ExtractedReference] = []
    _collect_from_struct(
        gff.root,
        "",
        file_type,
        script_fields,
        tag_fields,
        template_fields,
        conversation_fields,
        out,
    )

    # DLG: also walk nested paths from DLG_SCRIPT_PATHS (e.g. EntryList[].RepliesList[].Active)
    if file_type == "DLG":
        for list_field, script_field in DLG_SCRIPT_PATHS:
            if list_field not in gff.root:
                continue
            list_val = gff.root[list_field]
            if not isinstance(list_val, GFFList):
                continue
            for idx, item in enumerate(list_val):
                if isinstance(item, GFFStruct) and script_field in item:
                    val = item[script_field]
                    s = _value_to_str(val)
                    if s:
                        out.append(
                            ExtractedReference(
                                "script",
                                s,
                                f"{list_field}[{idx}].{script_field}",
                            ),
                        )

    return out


def find_referrers(
    installation: Installation,
    value: str,
    *,
    reference_kind: str = "resref",
    module_root: str | None = None,
    partial_match: bool = False,
    case_sensitive: bool = False,
    file_pattern: str | None = None,
    file_types: set[str] | None = None,
) -> list[ReferenceSearchResult]:
    """Find resources that reference the given value (script, tag, conversation, or resref).

    Uses reference_finder and optionally filters by module_root (e.g. "danm13" to limit
    to that module's RIMs). locations filter is not yet supported (installation iterator
    does not expose source location).

    Args:
    ----
        installation: Installation to search
        value: Script resref, tag, conversation resref, or generic resref
        reference_kind: "script" | "tag" | "conversation" | "resref"
        module_root: If set, only include results whose resource path contains this (e.g. module name)
        partial_match: Allow partial matches
        case_sensitive: Case-sensitive match
        file_pattern: Optional glob pattern for resource filenames
        file_types: Optional set of file type abbreviations

    Returns:
    -------
        List of ReferenceSearchResult; may be filtered by module_root
    """
    if reference_kind == "script":
        results = find_script_references(
            installation,
            value,
            partial_match=partial_match,
            case_sensitive=case_sensitive,
            file_pattern=file_pattern,
            file_types=file_types,
        )
    elif reference_kind == "tag":
        results = find_tag_references(
            installation,
            value,
            partial_match=partial_match,
            case_sensitive=case_sensitive,
            file_pattern=file_pattern,
            file_types=file_types,
        )
    elif reference_kind == "conversation":
        results = find_conversation_references(
            installation,
            value,
            partial_match=partial_match,
            case_sensitive=case_sensitive,
            file_pattern=file_pattern,
            file_types=file_types,
        )
    else:
        results = find_resref_references(
            installation,
            value,
            partial_match=partial_match,
            case_sensitive=case_sensitive,
            file_pattern=file_pattern,
            file_types=file_types,
        )

    if not module_root:
        return results

    key = module_root.lower()
    return [r for r in results if key in str(r.file_resource.filepath()).lower()]



from __future__ import annotations

from typing import TYPE_CHECKING

from qtpy.QtCore import Qt
from qtpy.QtGui import QFont
from qtpy.QtWidgets import QDialog, QHBoxLayout, QPushButton, QVBoxLayout, QWidget

from utility.gui.qt.widgets.font.font_widget import FontWidget

if TYPE_CHECKING:
    from qtpy.QtWidgets import QLayout


class FontDialog(QDialog):
    def __init__(
        self,
        font: QFont = QFont("Arial", 10),
        title: str = "Font",
    ):
        super().__init__()
        self._current_font: QFont = font
        self._init_ui(font=font, title=title)

    def _init_ui(
        self,
        font: QFont,
        title: str,
    ):
        self.setWindowTitle(title)
        self.setWindowFlags(Qt.WindowType.WindowCloseButtonHint)

        self._font_widget = FontWidget(font=font)
        font_widget_layout: QLayout | None = self._font_widget.layout()
        assert font_widget_layout is not None
        font_widget_layout.setContentsMargins(0, 0, 0, 0)

        ok_button = QPushButton("OK")
        cancel_button = QPushButton("Cancel")

        ok_button.clicked.connect(lambda: self.accept())
        cancel_button.clicked.connect(lambda: self.close())

        lay = QHBoxLayout()
        lay.setAlignment(Qt.AlignmentFlag.AlignRight)
        lay.addWidget(ok_button)
        lay.addWidget(cancel_button)
        lay.setContentsMargins(0, 2, 0, 0)

        ok_cancel_widget = QWidget()
        ok_cancel_widget.setLayout(lay)

        lay = QVBoxLayout()
        lay.addWidget(self._font_widget)
        lay.addWidget(ok_cancel_widget)
        self.setLayout(lay)

    def get_font(self) -> QFont:
        return self._font_widget.get_font()

    def get_font_widget(self) -> FontWidget:
        return self._font_widget



"""Core engine stubs used by tests.

This module provides minimal game engine infrastructure for testing purposes.

References:
----------

    Derivations and Other Implementations:
    -------------------------------------
        https://github.com/th3w1zard1/KotOR.js/tree/master/src/engine/Engine.ts


    Note: This is a minimal stub; full engine is in ./Engines directory
"""

from __future__ import annotations

from dataclasses import dataclass
from pathlib import Path
from typing import Dict, Iterable, Type, TypeVar

from .graphics import GUIManager

T = TypeVar("T")


class ServiceContainer:
    """Simple type-aware service registry."""

    def __init__(self, services: Iterable[object] | None = None) -> None:
        self._services: Dict[Type[object], object] = {}
        if services is not None:
            for service in services:
                self.register(service)

    def register(self, service: object) -> None:
        """Register a service instance keyed by its type."""
        self._services[type(service)] = service

    def get(self, service_type: Type[T]) -> T:
        """Retrieve a registered service instance."""
        service = self._services.get(service_type)
        if service is None:
            raise KeyError(f"No service registered for {service_type!r}")
        return service  # type: ignore[return-value]


@dataclass(frozen=True)
class EngineState:
    """Represents high-level engine state information."""

    game_path: Path
    exit_requested: bool = False


class KotorEngine:
    """Minimal stand-in for the real engine used in tests."""

    def __init__(self, game_path: Path) -> None:
        self._state = EngineState(game_path=Path(game_path))
        self.services = ServiceContainer(services=[GUIManager()])

    @property
    def game_path(self) -> Path:
        """Path to the KotOR installation."""
        return self._state.game_path

    def request_exit(self) -> None:
        """Request shutdown of the engine."""
        self._state = EngineState(game_path=self._state.game_path, exit_requested=True)



from __future__ import annotations

from enum import Enum
from typing import Any

import qtpy


def sip_enum_to_int(obj: Any) -> int:
    # First check if it's already an int
    if isinstance(obj, int):
        return obj
    # Check if it's a Python Enum (has value attribute)
    if isinstance(obj, Enum):
        return sip_enum_to_int(obj.value)
    # Check if it has a value attribute (Qt6 enum or other enum-like objects)
    # Try to access value attribute directly - if it exists, use it
    try:
        value_attr = obj.value  # type: ignore[attr-defined]
        return sip_enum_to_int(value_attr)
    except AttributeError:
        # Object doesn't have value attribute, continue to other checks
        pass
    # For Qt5 SIP enums, try to convert directly
    if qtpy.QT5:
        return int(obj)
    # Last resort: try to convert to int
    return int(obj)



"""Placeholder scene object (no geometry) for grouping or transform anchors."""

from __future__ import annotations

from typing import TYPE_CHECKING

if TYPE_CHECKING:
    from glm import mat4

    from pykotor.gl.scene import Scene
    from pykotor.gl.shader import Shader


class Empty:
    def __init__(self, scene: Scene):
        self._scene: Scene = scene

    def draw(self, shader: Shader, transform: mat4): ...



"""This module holds classes relating to read and write operations."""

from __future__ import annotations

import io
import mmap
import os
import struct

from abc import ABC, abstractmethod
from pathlib import Path
from typing import TYPE_CHECKING, cast

from loggerplus import RobustLogger
from pykotor.common.language import LocalizedString

# decode_bytes_with_fallbacks is imported lazily in the function that uses it to avoid circular imports
from utility.common.geometry import Vector2, Vector3, Vector4

if TYPE_CHECKING:
    from types import TracebackType

    from typing_extensions import Literal, Self  # pyright: ignore[reportMissingModuleSource]

    from pykotor.resource.type import SOURCE_TYPES, TARGET_TYPES


def _endian_char(
    big: bool,  # noqa: FBT001
) -> str:
    """Returns the character that represents either big endian or small endian in struct unpack.

    Args:
    ----
        big: True if big endian.

    Returns:
    -------
        Character representing either big or small endian.
    """
    return ">" if big else "<"


class ArrayHead:
    def __init__(
        self,
        array_offset: int = 0,
        array_length: int = 0,
    ):
        self.length: int = array_length
        self.offset: int = array_offset


class RawBinaryReader:
    """Provides easier reading of binary objects that abstracts uniformly to all different stream/data types."""

    def __init__(
        self,
        stream: io.RawIOBase | io.BufferedIOBase | mmap.mmap,
        offset: int = 0,
        size: int | None = None,
    ):
        self.auto_close: bool = True

        self._stream: io.RawIOBase | io.BufferedIOBase | mmap.mmap = stream
        self._offset: int = offset
        self._stream.seek(offset)

        total_size = self.true_size()
        if self._offset > total_size - (size or 0):
            msg = "Specified offset/size is greater than the number of available bytes."
            raise OSError(msg)
        if size and size < 0:
            msg = f"Size must be greater than zero, got {size}"
            raise ValueError(msg)

        self._size: int = total_size - self._offset if size is None else size

        self._position: int = 0
        if hasattr(self._stream, "tell"):
            try:
                self._position = self._stream.tell() - self._offset
            except (OSError, ValueError):
                self._position = 0

    @property
    def _true_stream_position(self) -> int:
        """Private property to access the current position of the stream for debugging purposes.

        Returns:
        -------
            int: The current absolute position of the stream pointer.
        """
        return self._stream.tell()

    def __enter__(self):
        return self

    def __exit__(
        self,
        exc_type: type[BaseException] | None,
        exc_val: BaseException | None,
        exc_tb: TracebackType | None,
    ):
        self.close()

    def read(self, size: int) -> bytes | None:
        data = self._stream.read(size)
        if data:
            self._position += len(data)
        return data

    @classmethod
    def from_stream(
        cls,
        stream: io.RawIOBase | io.BufferedIOBase,
        offset: int = 0,
        size: int | None = None,
    ) -> Self:
        if not stream.seekable():
            msg = "Stream must be seekable"
            raise ValueError(msg)
        try:
            mmap_stream = mmap.mmap(stream.fileno(), length=0, access=mmap.ACCESS_READ)
        except (ValueError, OSError):  # ValueError means mmap cannot map to empty files
            if isinstance(stream, io.RawIOBase):
                return cls(io.BufferedReader(stream), offset, size)
            return cls(stream, offset, size)
        else:
            return cls(mmap_stream, offset, size)

    @classmethod
    def from_file(
        cls,
        path: os.PathLike | str,
        offset: int = 0,
        size: int | None = None,
    ) -> Self:
        """Returns a new BinaryReader with a stream established to the specified path.

        Args:
        ----
            path: str or pathlike object of the file to open.
            offset: Number of bytes into the stream to consider as position 0.
            size: Number of bytes to allowed to read from the stream. If not specified, uses the whole stream.

        Returns:
        -------
            A new BinaryReader instance.
        """
        stream = open(path, "rb")  # noqa: SIM115, PTH123
        instance = cls.from_stream(stream, offset, size)
        if instance._stream is not stream:  # noqa: SLF001
            stream.close()
        return instance

    @classmethod
    def from_bytes(
        cls,
        data: bytes | memoryview | bytearray,
        offset: int = 0,
        size: int | None = None,
    ) -> Self:
        """Returns a new BinaryReader with a stream established to the bytes stored in memory.

        Args:
        ----
            data: The bytes of data.
            offset: Number of bytes into the stream to consider as position 0.
            size: Number of bytes to allowed to read from the stream. If not specified, uses the whole stream.

        Returns:
        -------
            A new BinaryReader instance.
        """
        stream = io.BytesIO(data)
        return cls(stream, offset, size)

    @classmethod
    def from_auto(
        cls,
        source: SOURCE_TYPES,
        offset: int = 0,
        size: int | None = None,
    ) -> Self:
        if isinstance(source, (os.PathLike, str)):  # is path
            reader = cls.from_file(source, offset, size)

        elif isinstance(source, (memoryview, bytes, bytearray)):  # is binary data
            reader = cls.from_bytes(source, offset, size)

        elif isinstance(source, (io.IOBase, mmap.mmap)):
            if isinstance(source, (io.RawIOBase, io.BufferedIOBase)):  # only seekable streams are supported.
                reader = cls.from_stream(source, offset, size)
            else:
                msg = f"Stream of type '{source.__class__}' is not supported by this {cls.__name__} class."
                raise TypeError(msg)

        elif isinstance(source, RawBinaryReader):  # is already a BinaryReader instance
            current_position = source.position()
            reader = cls(source.get_stream(), source.offset(), source.size())  # noqa: SLF001
            reader.seek(current_position)

        else:
            msg = f"Must specify a path, bytes-like object, stream, io. or an existing BinaryReader instance, got type ({source.__class__})."
            raise TypeError(msg)

        return reader

    @staticmethod
    def load_file(
        path: os.PathLike | str,
        offset: int = 0,
        size: int = -1,
    ) -> bytes:
        """Returns bytes of a file at from specified path.

        Args:
        ----
            path: The str or path-like path of the file.
            offset: The offset into the file.
            size: The amount of bytes to load, if size equals -1 loads the whole file.

        Returns:
        -------
            The bytes of the file.
        """
        with Path(path).open("rb") as reader:
            reader.seek(offset)
            return reader.read() if size == -1 else reader.read(size)

    def offset(self) -> int:
        """Returns the offset value.

        Args:
        ----
            self: The class instance

        Returns:
        -------
            int: The offset value
        """
        return self._offset

    def set_offset(
        self,
        offset: int,
    ):
        absolute_position: int = self._offset + self._position
        self.seek(self.position() + offset)
        self._offset = offset
        self._position = absolute_position

    def size(self) -> int:
        """Returns the total number of bytes in the stream.

        When a BinaryReader is instantiated, it can be given a size argument and an offset argument, and will not
        exceed those limits. Use true_size if you want to know this information.

        Returns:
        -------
            The number of accessible bytes.
        """
        return self._size

    def true_size(self) -> int:
        """Returns the total number of bytes in the stream.

        This does NOT include the initial offset and size constraint passed to the constructor.

        Returns:
        -------
            The absolute size of the stream.
        """
        if isinstance(self._stream, mmap.mmap):
            return self._stream.size()

        current: int = self._stream.tell()
        self._stream.seek(0, os.SEEK_END)
        size: int = self._stream.tell()
        self._stream.seek(current)
        return size

    def remaining(self) -> int:
        """Returns the number of bytes remaining in the stream based on current position.

        Returns:
        -------
            The total file size.
        """
        return self.size() - self.position()

    def close(self):
        """Closes the underlying stream and releases any resources."""
        self._stream.close()

    def skip(
        self,
        length: int,
    ):
        """Skips ahead in the stream the specified number of bytes.

        Args:
        ----
            length: How many bytes to skip.
        """
        self.exceed_check(length)
        skipped = self._stream.read(length) or b""
        self._position += len(skipped)

    def position(self) -> int:
        """Returns the byte offset into the stream.

        Returns:
        -------
            The byte offset.
        """
        return self._position

    def get_stream(self) -> io.RawIOBase | io.BufferedIOBase | mmap.mmap:
        """Returns the underlying stream for internal use.

        Returns:
        -------
            The underlying stream object.
        """
        return self._stream

    def get_position(self) -> int:
        """Returns the cached position for internal use.

        Returns:
        -------
            The cached position value.
        """
        return self._position

    def seek(
        self,
        position: int,
    ):
        """Moves the stream pointer to the byte offset.

        Args:
        ----
            position: The byte index into stream.
        """
        self.exceed_check(position - self._position)
        self._stream.seek(position + self._offset)
        self._position = position

    def read_all(self) -> bytes:
        """Read all remaining bytes from the stream.

        Args:
        ----
            self: The stream object

        Returns:
        -------
            bytes: The bytes read from the stream
        """
        remaining_bytes = self._size - self._position
        data = self._stream.read(remaining_bytes) or b""
        self._position += len(data)
        return data

    def _read_integer(
        self,
        size: int,
        format_char: str,
        *,
        big: bool = False,
    ) -> int:
        """Generic method to read an integer of specified size and format.

        Args:
        ----
            size: Number of bytes to read
            format_char: Struct format character (e.g., 'B', 'b', 'H', 'h', etc.)
            big: Read bytes as big endian

        Returns:
        -------
            The unpacked integer value
        """
        self.exceed_check(size)
        data = self._stream.read(size) or b""
        self._position += len(data)
        return struct.unpack(f"{_endian_char(big)}{format_char}", data)[0]

    def _read_float(
        self,
        size: int,
        format_char: str,
        *,
        big: bool = False,
    ) -> float:
        """Generic method to read a floating point number of specified size and format.

        Args:
        ----
            size: Number of bytes to read
            format_char: Struct format character (e.g., 'f', 'd')
            big: Read bytes as big endian

        Returns:
        -------
            The unpacked float value
        """
        self.exceed_check(size)
        data = self._stream.read(size) or b""
        self._position += len(data)
        return struct.unpack(f"{_endian_char(big)}{format_char}", data)[0]

    def read_uint8(
        self,
        *,
        big: bool = False,
    ) -> int:
        """Reads an unsigned 8-bit integer from the stream.

        Args:
        ----
            big: Read int bytes as big endian.

        Returns:
        -------
            An integer from the stream.
        """
        return self._read_integer(1, "B", big=big)

    def read_int8(
        self,
        *,
        big: bool = False,
    ) -> int:
        """Reads a signed 8-bit integer from the stream.

        Args:
        ----
            big: Read int bytes as big endian.

        Returns:
        -------
            An integer from the stream.
        """
        return self._read_integer(1, "b", big=big)

    def read_uint16(
        self,
        *,
        big: bool = False,
    ) -> int:
        """Reads an unsigned 16-bit integer from the stream.

        Args:
        ----
            big: Read int bytes as big endian.

        Returns:
        -------
            An integer from the stream.
        """
        return self._read_integer(2, "H", big=big)

    def read_int16(
        self,
        *,
        big: bool = False,
    ) -> int:
        """Reads a signed 16-bit integer from the stream.

        Args:
        ----
            big: Read int bytes as big endian.

        Returns:
        -------
            An integer from the stream.
        """
        return self._read_integer(2, "h", big=big)

    def read_uint32(
        self,
        *,
        max_neg1: bool = False,
        big: bool = False,
    ) -> int:
        """Reads an unsigned 32-bit integer from the stream.

        If max_is_neg1 flag is set to true and the bytes read off the stream are equal to 0xFFFFFFFF then the method
        will return a value of -1 instead of 4294967295 (hex 0xFFFFFFFF).

        Args:
        ----
            max_neg1: Return -1 when the value of the stream equals 0xFFFFFFFF (dec 4294967295).
            big: Read int bytes as big endian.

        Returns:
        -------
            An integer from the stream.
        """
        self.exceed_check(4)
        data = self._stream.read(4) or b""
        self._position += len(data)
        unpacked = struct.unpack(f"{_endian_char(big)}I", data)[0]

        if unpacked == 0xFFFFFFFF and max_neg1:  # noqa: PLR2004
            unpacked = -1

        return unpacked

    def read_int32(
        self,
        *,
        big: bool = False,
    ) -> int:
        """Reads a signed 32-bit integer from the stream.

        Args:
        ----
            big: Read int bytes as big endian.

        Returns:
        -------
            An integer from the stream.
        """
        return self._read_integer(4, "i", big=big)

    def read_uint64(
        self,
        *,
        big: bool = False,
    ) -> int:
        """Reads an unsigned 64-bit integer from the stream.

        Args:
        ----
            big: Read int bytes as big endian.

        Returns:
        -------
            An integer from the stream.
        """
        return self._read_integer(8, "Q", big=big)

    def read_int64(
        self,
        *,
        big: bool = False,
    ) -> int:
        """Reads a signed 64-bit integer from the stream.

        Args:
        ----
            big: Read int bytes as big endian.

        Returns:
        -------
            An integer from the stream.
        """
        return self._read_integer(8, "q", big=big)

    def read_single(
        self,
        *,
        big: bool = False,
    ) -> float:
        """Reads a 32-bit floating point number from the stream.

        Args:
        ----
            big: Read float bytes as big endian.

        Returns:
        -------
            A float from the stream.
        """
        return self._read_float(4, "f", big=big)

    def read_double(
        self,
        *,
        big: bool = False,
    ) -> float:
        """Reads a 64-bit floating point number from the stream.

        Args:
        ----
            big: Read float bytes as big endian.

        Returns:
        -------
            A float from the stream.
        """
        return self._read_float(8, "d", big=big)

    def read_vector2(
        self,
        *,
        big: bool = False,
    ) -> Vector2:
        """Reads two 32-bit floating point numbers from the stream.

        Args:
        ----
            big: Read bytes as big endian.

        Returns:
        -------
            A new Vector2 instance using floats read from the stream.
        """
        self.exceed_check(8)
        x, y = self.read_single(big=big), self.read_single(big=big)
        return Vector2(x, y)

    def read_vector3(
        self,
        *,
        big: bool = False,
    ) -> Vector3:
        """Reads three 32-bit floating point numbers from the stream.

        Args:
        ----
            big: Read bytes as big endian.

        Returns:
        -------
            A new Vector3 instance using floats read from the stream.
        """
        self.exceed_check(12)
        x, y, z = (
            self.read_single(big=big),
            self.read_single(big=big),
            self.read_single(big=big),
        )
        return Vector3(x, y, z)

    def read_vector4(
        self,
        *,
        big: bool = False,
    ) -> Vector4:
        """Reads four 32-bit floating point numbers from the stream.

        Args:
        ----
            big: Read bytes as big endian.

        Returns:
        -------
            A new Vector4 instance using floats read from the stream.
        """
        self.exceed_check(16)
        x, y, z, w = (
            self.read_single(big=big),
            self.read_single(big=big),
            self.read_single(big=big),
            self.read_single(big=big),
        )
        return Vector4(x, y, z, w)

    def read_line(
        self,
        encoding: str = "ascii",
        errors: str = "ignore",
    ) -> str:
        """Reads a line from the stream up to a line ending character.

        Args:
        ----
            encoding: The encoding to use for decoding the line.
            errors: How to handle encoding errors.

        Returns:
        -------
            A string representing the line read from the stream.
        """
        line_bytes = bytearray()
        while True:
            char: bytes | None = self._stream.read(1)
            if char:
                self._position += len(char)

            if not char:  # End of stream
                break

            line_bytes.extend(char)

            # Check for line endings
            if char == b"\n":
                break

            if char == b"\r":
                # Check for \r\n
                next_char: bytes | None = self._stream.read(1)
                if next_char:
                    self._position += len(next_char)
                if not next_char or next_char == b"\n":
                    break

                # If not \r\n, go back one character
                self._stream.seek(-1, os.SEEK_CUR)
                self._position -= 1
                break

        return bytes(line_bytes).decode(encoding=encoding, errors=errors).rstrip("\r\n")

    def read_bytes(
        self,
        length: int,
    ) -> bytes:
        """Reads a specified number of bytes from the stream.

        Args:
        ----
            length: Number of bytes to read.

        Returns:
        -------
            A bytes object containing the read bytes.
        """
        self.exceed_check(length)
        data = self._stream.read(length) or b""
        self._position += len(data)
        return data

    def read_string(
        self,
        length: int,
        encoding: str | None = "windows-1252",
        errors: Literal["ignore", "strict", "replace"] = "ignore",
    ) -> str:
        """Reads a string from the stream with the specified length.

        Any null bytes and characters proceeding a null byte are trimmed from the
        final value and any unknown characters are ignored.

        Args:
        ----
            length: Amount of character to read.
            encoding: Encoding of string to read.  If not specified, will default to 'windows-1252'. If set to None, will autodetect using charset_normalizer.

        Returns:
        -------
            A string read from the stream.
        """
        self.exceed_check(length)
        string_byte_data: bytes = self._stream.read(length) or b""
        self._position += len(string_byte_data)
        if encoding is None:
            # Lazy import to avoid circular dependency
            from pykotor.tools.encoding import decode_bytes_with_fallbacks

            string: str = decode_bytes_with_fallbacks(string_byte_data, encoding=encoding, errors=errors)
            RobustLogger().warning(f"decode_bytes_with_fallbacks called and returned '{string}'")
        else:
            string = string_byte_data.decode(encoding=encoding, errors=errors)
        if "\0" in string:
            string = string[: string.index("\0")].rstrip("\0")
            string = string.replace("\0", "")
        return string

    def read_terminated_string(
        self,
        terminator: str = "\0",
        length: int = -1,
        encoding: str = "ascii",
        *,
        strict: bool = True,
    ) -> str:
        """Reads a string continuously from the stream up to a specified length or until it hits the terminator character, whichever comes first.
        If length is -1, reads until the terminator is encountered without a length constraint.

        Args:
        ----
            terminator: The terminator character.
            length: The maximum length to read from the stream, or -1 for no length constraint.
                If a terminator is found before length is reached, skip the remaining.
            encoding: the encoding to use, in most cases this should be "ascii" (default)
            strict: Whether to stop appending the final string if a character could not be decoded by the encoding.

        Returns:
        -------
            A string read from the stream.
        """
        string: str = ""
        char: str = ""
        bytes_read: int = 0

        while char != terminator and (length == -1 or bytes_read < length):
            string += char
            self.exceed_check(1)
            char = self.read_bytes(1).decode(encoding=encoding, errors="ignore")
            bytes_read += 1
            if not char and strict:
                break

        if length == -1:
            return string

        # If a length is specified and not all bytes were read, skip the remaining bytes
        # Only skip if we haven't exceeded the stream boundaries
        remaining_length: int = length - bytes_read
        if remaining_length > 0:
            # Check if we can skip without exceeding boundaries
            if self._position + remaining_length <= self._size:
                self.skip(remaining_length)
            else:
                # Skip only what's available
                available = self._size - self._position
                if available > 0:
                    self.skip(available)
        return string

    def read_locstring(self) -> LocalizedString:
        """Reads the localized string data structure from the stream.

        The binary data structure that is read follows the structure found in the GFF format specification.

        Returns:
        -------
            A LocalizedString read from the stream.
        """
        locstring: LocalizedString = LocalizedString.from_invalid()
        self.skip(4)  # total number of bytes of the localized string
        locstring.stringref = self.read_uint32(max_neg1=True)
        string_count: int = self.read_uint32()
        for _ in range(string_count):
            string_id: int = self.read_uint32()
            language, gender = LocalizedString.substring_pair(string_id)
            length: int = self.read_uint32()
            string: str = self.read_string(length, encoding=language.get_encoding())
            locstring.set_data(language, gender, string)
        return locstring

    def read_array_head(self) -> ArrayHead:
        return ArrayHead(self.read_uint32(), self.read_uint32())

    def peek(
        self,
        length: int = 1,
    ) -> bytes:
        data: bytes | None = self._stream.read(length)
        self._stream.seek(-length, os.SEEK_CUR)
        return b"" if data is None else data

    def exceed_check(
        self,
        num: int,
    ):
        """Raises an error if the specified number exceeds the number of remaining bytes.

        Args:
        ----
            num: Number of bytes.

        Raises:
        ------
            OSError: When the attempted read operation exceeds the number of remaining bytes.
        """
        attempted_seek: int = self._position + num
        if attempted_seek < 0:
            msg = f"Cannot seek to a negative value {attempted_seek}, abstracted seek value: {num}"
            raise OSError(msg)
        if attempted_seek > self._size:
            msg = "This operation would exceed the streams boundaries."
            raise OSError(msg)


class RawBinaryWriter(ABC):
    def __enter__(self) -> Self:
        return self

    def __exit__(
        self,
        exc_type: type[BaseException] | None,
        exc_val: BaseException | None,
        exc_tb: TracebackType | None,
    ): ...

    @classmethod
    def to_file(
        cls,
        path: os.PathLike | str,
    ) -> RawBinaryWriterFile:
        """Returns a new BinaryWriter with a stream established to the specified path.

        Args:
        ----
            path: str or PathLike object of the file to open.

        Returns:
        -------
            A new BinaryWriter instance.
        """
        return RawBinaryWriterFile(Path(path).open("wb"))  # noqa: SIM115

    @classmethod
    def to_bytearray(
        cls,
        data: bytearray | None = None,
    ) -> RawBinaryWriterBytearray:
        """Returns a new BinaryWriter with a stream established to the specified bytes.

        Args:
        ----
            data: The bytes to write to.

        Returns:
        -------
            A new BinaryWriter instance.
        """
        if isinstance(data, (bytes, memoryview)):
            msg = "Must be bytearray, not bytes or memoryview."
            raise TypeError(msg)
        if data is None:
            data = bytearray()
        return RawBinaryWriterBytearray(data)

    @classmethod
    def to_stream(
        cls,
        stream: io.BufferedIOBase | io.RawIOBase,
        offset: int = 0,
    ) -> RawBinaryWriterFile:
        """Returns a new BinaryWriter backed by the given binary stream.

        Args:
        ----
            stream: The binary stream to write to.
            offset: Byte offset within the stream to begin writing from.

        Returns:
        -------
            A new BinaryWriter instance.
        """
        if not isinstance(stream, (io.BufferedIOBase, io.RawIOBase)):
            msg = "Stream must be a buffered or raw binary IO object."
            raise TypeError(msg)
        validated_stream = cast("io.BufferedIOBase | io.RawIOBase", stream)
        return RawBinaryWriterFile(validated_stream, offset)

    @classmethod
    def to_auto(
        cls,
        source: TARGET_TYPES,
    ) -> RawBinaryWriter:
        if isinstance(source, (os.PathLike, str)):  # is path
            return cls.to_file(source)
        if isinstance(source, bytearray):  # is mutable binary data
            return cls.to_bytearray(source)
        if isinstance(source, (bytes, memoryview)):  # is immutable binary data
            return cls.to_bytearray(bytearray(source))
        if isinstance(source, RawBinaryWriterFile):
            return RawBinaryWriterFile(source._stream, source._offset)  # noqa: SLF001
        if isinstance(source, RawBinaryWriterBytearray):
            return RawBinaryWriterBytearray(source._ba, source._offset)  # noqa: SLF001
        if isinstance(source, (io.RawIOBase, io.BufferedIOBase, mmap.mmap)):
            return RawBinaryWriterFile(source, 0)
        msg = f"Must specify a path, bytes object or an existing BinaryWriter instance. Instead got: {source.__class__.__name__}"
        raise NotImplementedError(msg)

    @staticmethod
    def dump(
        path: os.PathLike | str,
        data: bytes | bytearray | memoryview | mmap.mmap,
    ):
        """Convenience method used to write the specified data to the specified file.

        Args:
        ----
            path: The filepath of the file.
            data: The data to write to the file.
        """
        with Path(path).open("wb") as file:
            file.write(data)

    def close(self):
        """Closes the stream."""

    @abstractmethod
    def size(self) -> int:
        """Returns the total file size.

        Returns:
        -------
            The total file size.
        """

    @abstractmethod
    def data(self) -> bytes:
        """Returns the full file data.

        Returns:
        -------
            The full file data.
        """

    @abstractmethod
    def clear(self):
        """Clears all the data in the file."""

    @abstractmethod
    def seek(
        self,
        position: int,
    ):
        """Moves the stream pointer to the byte offset.

        Args:
        ----
            position: The byte index into stream.
        """

    @abstractmethod
    def end(self):
        """Moves the pointer for the stream to the end."""

    @abstractmethod
    def position(self) -> int:
        """Returns the byte offset into the stream.

        Returns:
        -------
            The byte offset.
        """

    @abstractmethod
    def write_uint8(
        self,
        value: int,
        *,
        big: bool = False,
    ):
        """Writes an unsigned 8-bit integer to the stream.

        Args:
        ----
            value: The value to be written.
            big: Write int bytes as big endian.
        """

    @abstractmethod
    def write_int8(
        self,
        value: int,
        *,
        big: bool = False,
    ):
        """Writes a signed 8-bit integer to the stream.

        Args:
        ----
            value: The value to be written.
            big: Write int bytes as big endian.
        """

    @abstractmethod
    def write_uint16(
        self,
        value: int,
        *,
        big: bool = False,
    ):
        """Writes an unsigned 16-bit integer to the stream.

        Args:
        ----
            value: The value to be written.
            big: Write int bytes as big endian.
        """

    @abstractmethod
    def write_int16(
        self,
        value: int,
        *,
        big: bool = False,
    ):
        """Writes a signed 16-bit integer to the stream.

        Args:
        ----
            value: The value to be written.
            big: Write int bytes as big endian.
        """

    @abstractmethod
    def write_uint32(
        self,
        value: int,
        *,
        max_neg1: bool = False,
        big: bool = False,
    ):
        """Writes an unsigned 32-bit integer to the stream.

        If the max_neg1 flag is set to true and the specified value is equal to -1 then the stream will accept the value
        and write 0xFFFFFFFF to the stream.

        Args:
        ----
            value: The value to be written.
            big: Write int bytes as big endian.
            max_neg1: When the value is -1 it is to be converted to the max uint32 value.
        """

    @abstractmethod
    def write_int32(
        self,
        value: int,
        *,
        big: bool = False,
    ):
        """Writes a signed 32-bit integer to the stream.

        Args:
        ----
            value: The value to be written.
            big: Write int bytes as big endian.
        """

    @abstractmethod
    def write_uint64(
        self,
        value: int,
        *,
        big: bool = False,
    ):
        """Writes an unsigned 64-bit integer to the stream.

        Args:
        ----
            value: The value to be written.
            big: Write int bytes as big endian.
        """

    @abstractmethod
    def write_int64(
        self,
        value: int,
        *,
        big: bool = False,
    ):
        """Writes a signed 64-bit integer to the stream.

        Args:
        ----
            value: The value to be written.
            big: Write int bytes as big endian.
        """

    @abstractmethod
    def write_single(
        self,
        value: float,
        *,
        big: bool = False,
    ):
        """Writes an 32-bit floating point number to the stream.

        Args:
        ----
            value: The value to be written.
            big: Write int bytes as big endian.
        """

    @abstractmethod
    def write_double(
        self,
        value: int,
        *,
        big: bool = False,
    ):
        """Writes a 64-bit floating point number to the stream.

        Args:
        ----
            value: The value to be written.
            big: Write bytes as big endian.
        """

    @abstractmethod
    def write_vector2(
        self,
        value: Vector2,
        *,
        big: bool = False,
    ):
        """Writes two 32-bit floating point numbers to the stream.

        Args:
        ----
            value: The value to be written.
            big: Write bytes as big endian.
        """

    @abstractmethod
    def write_vector3(
        self,
        value: Vector3,
        *,
        big: bool = False,
    ):
        """Writes three 32-bit floating point numbers to the stream.

        Args:
        ----
            value: The value to be written.
            big: Write bytes as big endian.
        """

    @abstractmethod
    def write_vector4(
        self,
        value: Vector4,
        *,
        big: bool = False,
    ):
        """Writes four 32-bit floating point numbers to the stream.

        Args:
        ----
            value: The value to be written.
            big: Write bytes as big endian.
        """

    @abstractmethod
    def write_bytes(
        self,
        value: bytes,
    ):
        """Writes the specified bytes to the stream.

        Args:
        ----
            value: The bytes to be written.
        """

    @abstractmethod
    def write_string(  # noqa: PLR0913
        self,
        value: str,
        encoding: str | None = "windows-1252",
        errors: str = "strict",
        *,
        big: bool = False,
        prefix_length: int = 0,
        string_length: int = -1,
        padding: str = "\0",
    ):
        """Writes the specified string to the stream.

        The string can also be prefixed by an integer specifying the strings length.

        Args:
        ----
            value: The string to be written.
            encoding: The string encoding. If not set, will use the default "windows-1252". If set to None, will autodetect using charset_normalizer.
            prefix_length: The number of bytes for the string length prefix. Valid options are 0, 1, 2 and 4.
            big: Write the prefix length integer as big endian.
            string_length: Fixes the string length to this size, truncating or padding where necessary. Ignores if -1.
            padding: What character is used as padding where applicable.
        """

    @abstractmethod
    def write_line(
        self,
        indent: int,
        *args,
    ):
        """Writes a line with specified indentation and array of values that are separated by whitespace.

        Args:
        ----
            indent: Level of indentation.
            *args: Values to write.
        """

    @abstractmethod
    def write_locstring(
        self,
        value: LocalizedString,
        *,
        big: bool = False,
    ):
        """Writes the specified localized string to the stream.

        The binary data structure that is read follows the structure found in the GFF format specification.

        Args:
        ----
            value: The localized string to be written.
            big: Write any integers as big endian.
        """


class RawBinaryWriterFile(RawBinaryWriter):
    def __init__(
        self,
        stream: io.BufferedIOBase | io.RawIOBase,
        offset: int = 0,
    ):
        self._stream: io.BufferedIOBase | io.RawIOBase = stream
        self._offset: int = offset
        self.auto_close: bool = True

        self._stream.seek(offset)

    def __exit__(
        self,
        exc_type: type[BaseException] | None,
        exc_val: BaseException | None,
        exc_tb: TracebackType | None,
    ):
        if self.auto_close:
            self.close()

    def close(self):
        """Closes the stream."""
        self._stream.close()

    def size(self) -> int:
        """Returns the total file size.

        Returns:
        -------
            The total file size.
        """
        pos: int = self._stream.tell()
        self._stream.seek(0, 2)
        size: int = self._stream.tell()
        self._stream.seek(pos)
        return size

    def data(self) -> bytes:
        """Returns the full file data.

        Returns:
        -------
            The full file data.
        """
        pos: int = self._stream.tell()
        self._stream.seek(0)
        data: bytes | None = self._stream.read()
        self._stream.seek(pos)
        return b"" if data is None else data

    def clear(self):
        """Clears all the data in the file."""
        self._stream.seek(0)
        self._stream.truncate()

    def seek(
        self,
        position: int,
    ):
        """Moves the stream pointer to the byte offset.

        Args:
        ----
            position: The byte index into stream.
        """
        self._stream.seek(position + self._offset)

    def end(self):
        """Moves the pointer for the stream to the end."""
        self._stream.seek(0, 2)

    def position(self) -> int:
        """Returns the byte offset into the stream.

        Returns:
        -------
            The byte offset.
        """
        return self._stream.tell() - self._offset

    def _write_integer(
        self,
        value: int,
        format_char: str,
        *,
        big: bool = False,
    ):
        """Generic method to write an integer with specified format.

        Args:
        ----
            value: The value to be written
            format_char: Struct format character (e.g., 'B', 'b', 'H', 'h', etc.)
            big: Write bytes as big endian
        """
        self._stream.write(struct.pack(f"{_endian_char(big)}{format_char}", value))

    def _write_float(
        self,
        value: float,
        format_char: str,
        *,
        big: bool = False,
    ):
        """Generic method to write a float with specified format.

        Args:
        ----
            value: The value to be written
            format_char: Struct format character (e.g., 'f', 'd')
            big: Write bytes as big endian
        """
        self._stream.write(struct.pack(f"{_endian_char(big)}{format_char}", value))

    def write_uint8(
        self,
        value: int,
        *,
        big: bool = False,
    ):
        """Writes an unsigned 8-bit integer to the stream.

        Args:
        ----
            value: The value to be written.
            big: Write int bytes as big endian.
        """
        self._write_integer(value, "B", big=big)

    def write_int8(
        self,
        value: int,
        *,
        big: bool = False,
    ):
        """Writes a signed 8-bit integer to the stream.

        Args:
        ----
            value: The value to be written.
            big: Write int bytes as big endian.
        """
        self._write_integer(value, "b", big=big)

    def write_uint16(
        self,
        value: int,
        *,
        big: bool = False,
    ):
        """Writes an unsigned 16-bit integer to the stream.

        Args:
        ----
            value: The value to be written.
            big: Write int bytes as big endian.
        """
        self._write_integer(value, "H", big=big)

    def write_int16(
        self,
        value: int,
        *,
        big: bool = False,
    ):
        """Writes a signed 16-bit integer to the stream.

        Args:
        ----
            value: The value to be written.
            big: Write int bytes as big endian.
        """
        self._write_integer(value, "h", big=big)

    def write_uint32(
        self,
        value: int,
        *,
        max_neg1: bool = False,
        big: bool = False,
    ):
        """Writes an unsigned 32-bit integer to the stream.

        If the max_neg1 flag is set to true and the specified value is equal to -1 then the stream will accept the value
        and write 0xFFFFFFFF to the stream.

        Args:
        ----
            value: The value to be written.
            big: Write int bytes as big endian.
            max_neg1: When the value is -1 it is to be converted to the max uint32 value.
        """
        if max_neg1 and value == -1:
            value = 0xFFFFFFFF

        self._stream.write(struct.pack(f"{_endian_char(big)}I", value))

    def write_int32(
        self,
        value: int,
        *,
        big: bool = False,
    ):
        """Writes a signed 32-bit integer to the stream.

        Args:
        ----
            value: The value to be written.
            big: Write int bytes as big endian.
        """
        self._write_integer(value, "i", big=big)

    def write_uint64(
        self,
        value: int,
        *,
        big: bool = False,
    ):
        """Writes an unsigned 64-bit integer to the stream.

        Args:
        ----
            value: The value to be written.
            big: Write int bytes as big endian.
        """
        self._write_integer(value, "Q", big=big)

    def write_int64(
        self,
        value: int,
        *,
        big: bool = False,
    ):
        """Writes a signed 64-bit integer to the stream.

        Args:
        ----
            value: The value to be written.
            big: Write int bytes as big endian.
        """
        self._write_integer(value, "q", big=big)

    def write_single(
        self,
        value: float,
        *,
        big: bool = False,
    ):
        """Writes a 32-bit floating point number to the stream.

        Args:
        ----
            value: The value to be written.
            big: Write bytes as big endian.
        """
        self._write_float(value, "f", big=big)

    def write_double(
        self,
        value: float,
        *,
        big: bool = False,
    ):
        """Writes a 64-bit floating point number to the stream.

        Args:
        ----
            value: The value to be written.
            big: Write bytes as big endian.
        """
        self._write_float(value, "d", big=big)

    def write_vector2(
        self,
        value: Vector2,
        *,
        big: bool = False,
    ):
        """Writes two 32-bit floating point numbers to the stream.

        Args:
        ----
            value: The value to be written.
            big: Write bytes as big endian.
        """
        self._stream.write(struct.pack(f"{_endian_char(big)}f", value.x))
        self._stream.write(struct.pack(f"{_endian_char(big)}f", value.y))

    def write_vector3(
        self,
        value: Vector3,
        *,
        big: bool = False,
    ):
        """Writes three 32-bit floating point numbers to the stream.

        Args:
        ----
            value: The value to be written.
            big: Write bytes as big endian.
        """
        self._stream.write(struct.pack(f"{_endian_char(big)}f", value.x))
        self._stream.write(struct.pack(f"{_endian_char(big)}f", value.y))
        self._stream.write(struct.pack(f"{_endian_char(big)}f", value.z))

    def write_vector4(
        self,
        value: Vector4,
        *,
        big: bool = False,
    ):
        """Writes four 32-bit floating point numbers to the stream.

        Args:
        ----
            value: The value to be written.
            big: Write bytes as big endian.
        """
        self._stream.write(struct.pack(f"{_endian_char(big)}f", value.x))
        self._stream.write(struct.pack(f"{_endian_char(big)}f", value.y))
        self._stream.write(struct.pack(f"{_endian_char(big)}f", value.z))
        self._stream.write(struct.pack(f"{_endian_char(big)}f", value.w))

    def write_bytes(
        self,
        value: bytes | bytearray,
    ):
        """Writes the specified bytes to the stream.

        Args:
        ----
            value: The bytes to be written.
        """
        self._stream.write(value)

    def write_string(  # noqa: PLR0913
        self,
        value: str,
        encoding: str | None = "windows-1252",
        errors: str = "strict",
        *,
        big: bool = False,
        prefix_length: int = 0,
        string_length: int = -1,
        padding: str = "\0",
    ):  # sourcery skip: inline-variable, switch
        """Writes the specified string to the stream.

        The string can also be prefixed by an integer specifying the strings length.

        Args:
        ----
            value: The string to be written.
            encoding: The encoding of the string to be written. If not set, will default to "windows-1252".
            prefix_length: The number of bytes for the string length prefix. Valid options are 0, 1, 2 and 4.
            big: Write the prefix length integer as big endian.
            string_length: Fixes the string length to this size, truncating or padding where necessary. Ignores if -1.
            padding: What character is used as padding where applicable.
        """
        if prefix_length == 0:
            pass
        elif prefix_length == 1:
            if len(value) > 0xFF:  # noqa: PLR2004
                msg = "The string length is too large for a prefix length of 1."
                raise ValueError(msg)
            self.write_uint8(len(value), big=big)

        elif prefix_length == 2:  # noqa: PLR2004
            if len(value) > 0xFFFF:  # noqa: PLR2004
                msg = "The string length is too large for a prefix length of 2."
                raise ValueError(msg)
            self.write_uint16(len(value), big=big)

        elif prefix_length == 4:  # noqa: PLR2004
            if len(value) > 0xFFFFFFFF:  # noqa: PLR2004
                msg = "The string length is too large for a prefix length of 4."
                raise ValueError(msg)
            self.write_uint32(len(value), big=big)

        else:
            msg = f"An invalid prefix length '{prefix_length}' was provided."
            raise ValueError(msg)

        if string_length != -1:
            while len(value) < string_length:
                value += padding
            value = value[:string_length]
        self._stream.write(value.encode(encoding or "windows-1252", errors=errors))

    def write_line(
        self,
        indent: int,
        *args: float | str,
    ):
        """Writes a line with specified indentation and array of values that are separated by whitespace.

        The data will be written with utf-8 encoding.

        Args:
        ----
            indent: Level of indentation.
            *args: Values to write.
        """
        line: str = "  " * indent
        for arg in args:
            line += str(round(arg, 7)) if isinstance(arg, float) else str(arg)
            line += " "
        line += "\n"
        self._stream.write(line.encode())

    def write_locstring(
        self,
        value: LocalizedString,
        *,
        big: bool = False,
    ):
        """Writes the specified localized string to the stream.

        The binary data structure that is read follows the structure found in the GFF format specification.

        Args:
        ----
            value: The localized string to be written.
            big: Write any integers as big endian.
        """
        bw: RawBinaryWriterBytearray = RawBinaryWriter.to_bytearray()
        bw.write_uint32(value.stringref, big=big, max_neg1=True)
        bw.write_uint32(len(value), big=big)

        for language, gender, substring in value:
            string_id: int = LocalizedString.substring_id(language, gender)
            bw.write_uint32(string_id, big=big)
            bw.write_string(substring, prefix_length=4, encoding=language.get_encoding())

        locstring_data: bytes = bw.data()
        self.write_uint32(len(locstring_data))
        self.write_bytes(locstring_data)


class RawBinaryWriterBytearray(RawBinaryWriter):
    def __init__(
        self,
        ba: bytes | bytearray,
        offset: int = 0,
    ):
        self._ba: bytearray = ba if isinstance(ba, bytearray) else bytearray(ba)
        self._offset: int = offset
        self._position: int = 0
        self._initial_size: int = len(ba)  # Track initial size to distinguish fixed-size from growable buffers

    def size(self) -> int:
        """Returns the total file size.

        Returns:
        -------
            The total file size.
        """
        return len(self._ba)

    def data(self) -> bytes:
        """Returns the full file data.

        Returns:
        -------
            The full file data.
        """
        return bytes(self._ba)

    def clear(self):
        """Clears all the data in the file."""
        self._ba.clear()

    def seek(
        self,
        position: int,
    ):
        """Moves the stream pointer to the byte offset.

        Args:
        ----
            position: The byte index into stream.
        """
        self._position = position

    def end(self):
        """Moves the pointer for the stream to the end."""
        self._position = len(self._ba)

    def position(self) -> int:
        """Returns the byte offset into the stream.

        Returns:
        -------
            The byte offset.
        """
        return self._position - self._offset

    def write_uint8(
        self,
        value: int,
        *,
        big: bool = False,
    ):
        """Writes an unsigned 8-bit integer to the stream.

        Args:
        ----
            value: The value to be written.
            big: Write int bytes as big endian.
        """
        # Only enforce bounds checking for fixed-size buffers (non-empty initial size)
        # Empty buffers are allowed to grow automatically
        if self._initial_size > 0 and self._position >= self._initial_size:
            raise IndexError(f"Cannot write beyond buffer bounds: position {self._position} >= initial size {self._initial_size}")
        self._ba[self._position : self._position + 1] = struct.pack(
            f"{_endian_char(big)}B",
            value,
        )
        self._position += 1

    def write_int8(
        self,
        value: int,
        *,
        big: bool = False,
    ):
        """Writes a signed 8-bit integer to the stream.

        Args:
        ----
            value: The value to be written.
            big: Write int bytes as big endian.
        """
        # Only enforce bounds checking for fixed-size buffers (non-empty initial size)
        # Empty buffers are allowed to grow automatically
        if self._initial_size > 0 and self._position >= self._initial_size:
            raise IndexError(f"Cannot write beyond buffer bounds: position {self._position} >= initial size {self._initial_size}")
        self._ba[self._position : self._position + 1] = struct.pack(
            f"{_endian_char(big)}b",
            value,
        )
        self._position += 1

    def write_uint16(
        self,
        value: int,
        *,
        big: bool = False,
    ):
        """Writes an unsigned 16-bit integer to the stream.

        Args:
        ----
            value: The value to be written.
            big: Write int bytes as big endian.
        """
        # Only enforce bounds checking for fixed-size buffers (non-empty initial size)
        # Empty buffers are allowed to grow automatically
        if self._initial_size > 0 and self._position + 2 > self._initial_size:
            raise IndexError(f"Cannot write beyond buffer bounds: position {self._position} + 2 > initial size {self._initial_size}")
        self._ba[self._position : self._position + 2] = struct.pack(
            f"{_endian_char(big)}H",
            value,
        )
        self._position += 2

    def write_int16(
        self,
        value: int,
        *,
        big: bool = False,
    ):
        """Writes a signed 16-bit integer to the stream.

        Args:
        ----
            value: The value to be written.
            big: Write int bytes as big endian.
        """
        # Only enforce bounds checking for fixed-size buffers (non-empty initial size)
        # Empty buffers are allowed to grow automatically
        if self._initial_size > 0 and self._position + 2 > self._initial_size:
            raise IndexError(f"Cannot write beyond buffer bounds: position {self._position} + 2 > initial size {self._initial_size}")
        self._ba[self._position : self._position + 2] = struct.pack(
            f"{_endian_char(big)}h",
            value,
        )
        self._position += 2

    def write_uint32(
        self,
        value: int,
        *,
        max_neg1: bool = False,
        big: bool = False,
    ):
        """Writes an unsigned 32-bit integer to the stream.

        If the max_neg1 flag is set to true and the specified value is equal to -1 then the stream will accept the value
        and write 0xFFFFFFFF to the stream.

        Args:
        ----
            value: The value to be written.
            big: Write int bytes as big endian.
            max_neg1: When the value is -1 it is to be converted to the max uint32 value.
        """
        if max_neg1 and value == -1:
            value = 0xFFFFFFFF
        # Only enforce bounds checking for fixed-size buffers (non-empty initial size)
        # Empty buffers are allowed to grow automatically
        if self._initial_size > 0 and self._position + 4 > self._initial_size:
            raise IndexError(f"Cannot write beyond buffer bounds: position {self._position} + 4 > initial size {self._initial_size}")
        self._ba[self._position : self._position + 4] = struct.pack(
            f"{_endian_char(big)}I",
            value,
        )
        self._position += 4

    def write_int32(
        self,
        value: int,
        *,
        big: bool = False,
    ):
        """Writes a signed 32-bit integer to the stream.

        Args:
        ----
            value: The value to be written.
            big: Write int bytes as big endian.
        """
        # Only enforce bounds checking for fixed-size buffers (non-empty initial size)
        # Empty buffers are allowed to grow automatically
        if self._initial_size > 0 and self._position + 4 > self._initial_size:
            raise IndexError(f"Cannot write beyond buffer bounds: position {self._position} + 4 > initial size {self._initial_size}")
        self._ba[self._position : self._position + 4] = struct.pack(
            f"{_endian_char(big)}i",
            value,
        )
        self._position += 4

    def write_uint64(
        self,
        value: int,
        *,
        big: bool = False,
    ):
        """Writes an unsigned 64-bit integer to the stream.

        Args:
        ----
            value: The value to be written.
            big: Write int bytes as big endian.
        """
        # Only enforce bounds checking for fixed-size buffers (non-empty initial size)
        # Empty buffers are allowed to grow automatically
        if self._initial_size > 0 and self._position + 8 > self._initial_size:
            raise IndexError(f"Cannot write beyond buffer bounds: position {self._position} + 8 > initial size {self._initial_size}")
        self._ba[self._position : self._position + 8] = struct.pack(
            f"{_endian_char(big)}Q",
            value,
        )
        self._position += 8

    def write_int64(
        self,
        value: int,
        *,
        big: bool = False,
    ):
        """Writes a signed 64-bit integer to the stream.

        Args:
        ----
            value: The value to be written.
            big: Write int bytes as big endian.
        """
        # Only enforce bounds checking for fixed-size buffers (non-empty initial size)
        # Empty buffers are allowed to grow automatically
        if self._initial_size > 0 and self._position + 8 > self._initial_size:
            raise IndexError(f"Cannot write beyond buffer bounds: position {self._position} + 8 > initial size {self._initial_size}")
        self._ba[self._position : self._position + 8] = struct.pack(
            f"{_endian_char(big)}q",
            value,
        )
        self._position += 8

    def write_single(
        self,
        value: float,
        *,
        big: bool = False,
    ):
        """Writes an 32-bit floating point number to the stream.

        Args:
        ----
            value: The value to be written.
            big: Write int bytes as big endian.
        """
        # Only enforce bounds checking for fixed-size buffers (non-empty initial size)
        # Empty buffers are allowed to grow automatically
        if self._initial_size > 0 and self._position + 4 > self._initial_size:
            raise IndexError(f"Cannot write beyond buffer bounds: position {self._position} + 4 > initial size {self._initial_size}")
        self._ba[self._position : self._position + 4] = struct.pack(
            f"{_endian_char(big)}f",
            value,
        )
        self._position += 4

    def write_double(
        self,
        value: int,
        *,
        big: bool = False,
    ):
        """Writes a 64-bit floating point number to the stream.

        Args:
        ----
            value: The value to be written.
            big: Write bytes as big endian.
        """
        # Only enforce bounds checking for fixed-size buffers (non-empty initial size)
        # Empty buffers are allowed to grow automatically
        if self._initial_size > 0 and self._position + 8 > self._initial_size:
            raise IndexError(f"Cannot write beyond buffer bounds: position {self._position} + 8 > initial size {self._initial_size}")
        self._ba[self._position : self._position + 8] = struct.pack(
            f"{_endian_char(big)}d",
            value,
        )
        self._position += 8

    def write_vector2(
        self,
        value: Vector2,
        *,
        big: bool = False,
    ):
        """Writes two 32-bit floating point numbers to the stream.

        Args:
        ----
            value: The value to be written.
            big: Write bytes as big endian.
        """
        self._ba[self._position : self._position + 4] = struct.pack(
            f"{_endian_char(big)}f",
            value.x,
        )
        self._ba[self._position + 4 : self._position + 8] = struct.pack(
            f"{_endian_char(big)}f",
            value.y,
        )
        self._position += 8

    def write_vector3(
        self,
        value: Vector3,
        *,
        big: bool = False,
    ):
        """Writes three 32-bit floating point numbers to the stream.

        Args:
        ----
            value: The value to be written.
            big: Write bytes as big endian.
        """
        self._ba[self._position : self._position + 4] = struct.pack(
            f"{_endian_char(big)}f",
            value.x,
        )
        self._ba[self._position + 4 : self._position + 8] = struct.pack(
            f"{_endian_char(big)}f",
            value.y,
        )
        self._ba[self._position + 8 : self._position + 12] = struct.pack(
            f"{_endian_char(big)}f",
            value.z,
        )
        self._position += 12

    def write_vector4(
        self,
        value: Vector4,
        *,
        big: bool = False,
    ):
        """Writes four 32-bit floating point numbers to the stream.

        Args:
        ----
            value: The value to be written.
            big: Write bytes as big endian.
        """
        self._ba[self._position : self._position + 4] = struct.pack(
            f"{_endian_char(big)}f",
            value.x,
        )
        self._ba[self._position + 4 : self._position + 8] = struct.pack(
            f"{_endian_char(big)}f",
            value.y,
        )
        self._ba[self._position + 8 : self._position + 12] = struct.pack(
            f"{_endian_char(big)}f",
            value.z,
        )
        self._ba[self._position + 12 : self._position + 16] = struct.pack(
            f"{_endian_char(big)}f",
            value.w,
        )
        self._position += 16

    def write_bytes(
        self,
        value: bytes | bytearray,
    ):
        """Writes the specified bytes to the stream.

        Args:
        ----
            value: The bytes to be written.
        """
        # Only enforce bounds checking for fixed-size buffers (non-empty initial size)
        # Empty buffers are allowed to grow automatically
        if self._initial_size > 0 and self._position + len(value) > self._initial_size:
            raise IndexError(f"Cannot write beyond buffer bounds: position {self._position} + {len(value)} > initial size {self._initial_size}")
        self._ba[self._position : self._position + len(value)] = value
        self._position += len(value)

    def write_string(  # noqa: PLR0913
        self,
        value: str,
        encoding: str | None = "windows-1252",
        errors: str = "strict",
        *,
        big: bool = False,
        prefix_length: int = 0,
        string_length: int = -1,
        padding: str = "\0",
    ):  # sourcery skip: inline-variable, switch
        """Writes the specified string to the stream.

        The string can also be prefixed by an integer specifying the strings length.

        Args:
        ----
            value: The string to be written.
            encoding: The encoding to convert to bytes. Defaults to "windows-1252".
            prefix_length: The number of bytes for the string length prefix. Valid options are 0, 1, 2 and 4.
            big: Write the prefix length integer as big endian.
            string_length: Fixes the string length to this size, truncating or padding where necessary. Ignores if -1.
            padding: What character is used as padding where applicable.
        """
        if prefix_length == 0:
            pass
        elif prefix_length == 1:
            if len(value) > 0xFF:  # noqa: PLR2004
                msg = "The string length is too large for a prefix length of 1."
                raise ValueError(msg)
            self.write_uint8(len(value), big=big)
        elif prefix_length == 2:  # noqa: PLR2004
            if len(value) > 0xFFFF:  # noqa: PLR2004
                msg = "The string length is too large for a prefix length of 2."
                raise ValueError(msg)
            self.write_uint16(len(value), big=big)
        elif prefix_length == 4:  # noqa: PLR2004
            if len(value) > 0xFFFFFFFF:  # noqa: PLR2004
                msg = "The string length is too large for a prefix length of 4."
                raise ValueError(msg)
            self.write_uint32(len(value), big=big)
        else:
            msg = "An invalid prefix length was provided."
            raise ValueError(msg)

        if string_length != -1:
            while len(value) < string_length:
                value += padding
            value = value[:string_length]

        self._encode_val_and_update_position(value, encoding, errors)

    def write_terminated_string(
        self,
        value: str,
        encoding: str | None = "windows-1252",
        errors: str = "strict",
    ):
        """Writes the specified string to the stream with a terminating null character."""
        encoded: bytes = value.encode(encoding or "windows-1252", errors=errors)
        required_size = self._position + len(encoded) + 1
        if len(self._ba) < required_size:
            self._ba.extend(b"\x00" * (required_size - len(self._ba)))
        self._ba[self._position : self._position + len(encoded)] = encoded
        self._ba[self._position + len(encoded)] = 0
        self._position += len(encoded) + 1

    def write_line(
        self,
        indent: int,
        *args: list[float] | list[str],
    ):
        """Writes a line with specified indentation and array of values that are separated by whitespace.

        Args:
        ----
            indent: Level of indentation.
            *args: Values to write.
        """
        line: str = "  " * indent
        for arg in args:
            line += str(round(arg, 7)) if isinstance(arg, float) else str(arg)
            line += " "
        line += "\n"

        self._encode_val_and_update_position(line, "ascii")

    def _encode_val_and_update_position(
        self,
        value: str,
        encoding: str | None,
        errors: str = "strict",
    ):
        encoded: bytes = value.encode(encoding or "windows-1252", errors=errors)
        self._ba[self._position : self._position + len(encoded)] = encoded
        self._position += len(encoded)

    def write_locstring(
        self,
        value: LocalizedString,
        *,
        big: bool = False,
    ):
        """Writes the specified localized string to the stream.

        The binary data structure that is read follows the structure found in the GFF format specification.

        Args:
        ----
            value: The localized string to be written.
            big: Write any integers as big endian.
        """
        bw: RawBinaryWriterBytearray = RawBinaryWriter.to_bytearray()
        bw.write_uint32(value.stringref, big=big, max_neg1=True)
        bw.write_uint32(len(value), big=big)

        for language, gender, substring in value:
            string_id: int = LocalizedString.substring_id(language, gender)
            bw.write_uint32(string_id, big=big)
            bw.write_string(substring, prefix_length=4, encoding=language.get_encoding(), errors="replace")

        locstring_data: bytes = bw.data()
        self.write_uint32(len(locstring_data))
        self.write_bytes(locstring_data)


if __name__ == "__main__":
    import random
    import time

    if TYPE_CHECKING:
        from typing_extensions import Self

    # Constants
    TEST_FILE = "test_file.bin"
    NUM_OPERATIONS = 10000
    NUM_INSTANTIATIONS = 10
    FILE_SIZE = 500 * 1024 * 1024  # 500MB
    FILE_DATA: bytes | None = None

    # Function to perform the I/O operations
    def test_io_performance(  # noqa: C901, PLR0912, PLR0915
        stream_class: type,
        mode: str = "rb",
    ) -> tuple[float, float, float]:
        print(f"Testing {stream_class.__name__}, mode={mode}")
        assert FILE_DATA is not None
        instantiation_times: list[float] = []
        operation_times: list[float] = []

        for i in range(NUM_INSTANTIATIONS):
            try:
                instantiation_start_time: float = time.time()
                if stream_class is RawBinaryReader:
                    if mode == "file":
                        stream = RawBinaryReader.from_file(TEST_FILE)
                    elif mode == "bytes":
                        instantiation_start_time = time.time()
                        stream = RawBinaryReader.from_bytes(FILE_DATA)
                    elif mode == "mmap":
                        raw_raw_stream = open(TEST_FILE, "rb")  # noqa: PTH123, SIM115
                        raw_stream = mmap.mmap(raw_raw_stream.fileno(), os.stat(TEST_FILE).st_size, access=mmap.ACCESS_READ)  # noqa: PTH116
                        instantiation_start_time = time.time()
                        stream = RawBinaryReader(raw_stream)
                    elif mode == "stream(io.BufferedReader)":
                        raw_raw_stream = open(TEST_FILE, "rb")  # noqa: PTH123, SIM115
                        raw_stream = io.BufferedReader(raw_raw_stream)  # pyright: ignore[reportArgumentType]
                        instantiation_start_time = time.time()
                        stream = RawBinaryReader.from_stream(raw_stream)
                    elif mode == "stream(io.BufferedRandom)":
                        raw_raw_stream = open(TEST_FILE, "r+b")  # noqa: PTH123, SIM115
                        raw_stream = io.BufferedRandom(raw_raw_stream)  # pyright: ignore[reportArgumentType]
                        instantiation_start_time = time.time()
                        stream = RawBinaryReader.from_stream(raw_stream)
                    elif mode == "stream(io.BytesIO)":
                        # Special handling for BytesIO
                        stream = io.BytesIO(FILE_DATA)
                        instantiation_start_time = time.time()
                        stream = RawBinaryReader.from_stream(stream)
                    elif mode == "stream(io.FileIO)":
                        raw_stream = io.FileIO(TEST_FILE, "rb")
                        instantiation_start_time = time.time()
                        stream = RawBinaryReader.from_stream(raw_stream)
                    elif mode == "stream(raw)":
                        raw_stream = open(TEST_FILE, "rb")  # noqa: PTH123, SIM115
                        instantiation_start_time = time.time()
                        stream = RawBinaryReader.from_stream(raw_stream)
                    else:
                        raise ValueError(f"cannot test mode: {mode}")
                elif stream_class is io.BytesIO:
                    # Special handling for BytesIO
                    stream = io.BytesIO(FILE_DATA)
                else:
                    raw_stream = open(TEST_FILE, mode)  # noqa: PTH123, SIM115
                    if stream_class is io.BufferedReader:
                        stream = io.BufferedReader(raw_stream)  # pyright: ignore[reportArgumentType]
                    elif stream_class is io.BufferedRandom:
                        stream = io.BufferedRandom(raw_stream)  # pyright: ignore[reportArgumentType]
                    else:
                        stream = stream_class(TEST_FILE, mode)  # pyright: ignore[reportArgumentType, reportCallIssue]
                instantiation_end_time: float = time.time()
                instantiation_times.append(instantiation_end_time - instantiation_start_time)
            finally:
                if i != NUM_INSTANTIATIONS - 1:
                    stream.close()

        try:
            operation_start_time: float = time.time()
            for _ in range(NUM_OPERATIONS):
                seek_position: int = random.randint(0, os.path.getsize(TEST_FILE) - 1)  # noqa: S311, PTH202
                stream.seek(seek_position)
                stream.read(1)
            operation_end_time: float = time.time()
            operation_times.append(operation_end_time - operation_start_time)
        finally:
            stream.close()

        total_instantiation_time: float = sum(instantiation_times)
        total_operation_time: float = sum(operation_times)
        total_time: float = total_instantiation_time + total_operation_time

        return total_instantiation_time, total_operation_time, total_time

    # Main function to run the tests
    def main():
        global FILE_DATA  # noqa: PLW0603

        results: list[list[int | float]] = []
        # Test using different stream types
        stream_types: list[tuple[type, str]] = [
            (io.FileIO, "rb"),  # Raw access
            (io.BytesIO, "rb"),  # In-memory stream (needs special handling)
            (io.BufferedReader, "rb"),  # Buffered reading
            (io.BufferedRandom, "r+b"),  # Buffered random access
            (RawBinaryReader, "file"),
            (RawBinaryReader, "mmap"),
            (RawBinaryReader, "stream(io.BufferedReader)"),
            (RawBinaryReader, "stream(io.BufferedRandom)"),
            (RawBinaryReader, "stream(io.BytesIO)"),
            (RawBinaryReader, "stream(io.FileIO)"),
            (RawBinaryReader, "stream(raw)"),
        ]

        # Ensure the test file exists and is the correct size
        if not os.path.exists(TEST_FILE) or os.path.getsize(TEST_FILE) != FILE_SIZE:  # noqa: PTH110, PTH202
            FILE_DATA = os.urandom(FILE_SIZE)
            print("Creating test file...")
            with open(TEST_FILE, "wb") as f:  # noqa: PTH123
                f.write(FILE_DATA)
        if FILE_DATA is None:
            with open(TEST_FILE, "rb") as f:  # noqa: PTH123
                FILE_DATA = f.read()

        # Run the tests
        for stream_class, mode in stream_types:
            print(f"Testing {stream_class.__name__}, mode={mode}")
            total_instantiation_time, total_operation_time, total_time = test_io_performance(stream_class, mode)
            results.append(
                [
                    f"{stream_class.__name__}({mode})",  # pyright: ignore[reportArgumentType]
                    total_instantiation_time,
                    total_operation_time,
                    total_time,
                ]
            )

        # Sort by total performance (fastest first)
        results.sort(key=lambda x: x[3])
        fastest_total_time: float = results[0][3]
        fastest_instantiation_time: float = min(results, key=lambda x: x[1])[1]
        fastest_operation_time: float = min(results, key=lambda x: x[2])[2]

        for result in results:
            for i, elem in enumerate(result):
                if elem != 0:
                    continue
                result[i] = 0.0001

        # Print the results with additional statistics
        print("------------------------------------------------------\n\nInstantiation Statistics (sorted by fastest to slowest):\n")
        for result in results:
            speed_percent: float = (fastest_instantiation_time / result[1]) * 100
            print(f"{result[0]}: {result[1]:.4f} seconds ({speed_percent:.2f}% of fastest, {result[1] / NUM_INSTANTIATIONS:.2f}s per)")

        print("------------------------------------------------------\n\nOperation Statistics (sorted by fastest to slowest):")
        for result in results:
            speed_percent = (fastest_operation_time / result[2]) * 100
            print(f"{result[0]}: {result[2]:.4f} seconds ({speed_percent:.2f}% of fastest, {result[2] / NUM_INSTANTIATIONS:.2f}s per)")

        print("------------------------------------------------------\n\nCombined Statistics (sorted by fastest to slowest):")
        for result in results:
            speed_percent = (fastest_total_time / result[3]) * 100
            print(f"{result[0]}: {result[3]:.4f} seconds (Inst: {result[1]:.4f}s, Ops: {result[2]:.4f}s) ({speed_percent:.2f}% of fastest)")
        print("------------------------------------------------------\n")

        # Calculate average times
        avg_instantiation_time: float = sum(r[1] for r in results) / len(results)
        avg_operation_time: float = sum(r[2] for r in results) / len(results)
        avg_total_time: float = sum(r[3] for r in results) / len(results)

        print(f"Average Instantiation Time: {avg_instantiation_time:.4f} seconds")
        print(f"Average Operation Time: {avg_operation_time:.4f} seconds")
        print(f"Average Total Time: {avg_total_time:.4f} seconds")
        print(f"Fastest Total Time: {fastest_total_time:.4f} seconds")
        print(f"Slowest Total Time: {results[-1][3]:.4f} seconds")
        print(f"Speed Ratio (Slowest/Fastest): {results[-1][3] / fastest_total_time:.2f}\n")

    main()



"""Bitmap font grid: cell coordinates and dimensions for texture atlases."""

from __future__ import annotations

import math

from dataclasses import dataclass
from typing import NamedTuple


class GridCell(NamedTuple):
    """Represents a cell in the bitmap font grid."""

    x: int
    y: int
    pixel_x1: float
    pixel_y1: float
    pixel_x2: float
    pixel_y2: float
    norm_x1: float
    norm_y1: float
    norm_x2: float
    norm_y2: float


@dataclass
class GridDimensions:
    """Stores the calculated dimensions for the bitmap font grid."""

    characters_per_row: int
    characters_per_column: int
    cell_size: int
    cell_height: float


class BitmapGrid:
    """Handles grid calculations for bitmap font generation."""

    def __init__(
        self,
        resolution: tuple[int, int],
        num_chars: int,
    ):
        """Initialize the bitmap grid calculator.

        Args:
            resolution: Tuple of (width, height) for the bitmap
            num_chars: Number of characters to fit in the grid

        Raises:
            ZeroDivisionError: If resolution contains zero values
        """
        if any(r == 0 for r in resolution):
            msg: str = f"resolution must be nonzero, got {resolution}"
            raise ZeroDivisionError(msg)

        self.resolution: tuple[int, int] = resolution
        self.num_chars: int = num_chars
        self.dimensions: GridDimensions = self._calculate_dimensions()

    def _calculate_dimensions(self) -> GridDimensions:
        """Calculate the grid dimensions based on resolution and character count.

        Returns:
            GridDimensions object containing the calculated dimensions
        """
        # Calculate grid layout
        chars_per_col: int = math.ceil(math.sqrt(self.num_chars))
        chars_per_row: int = math.ceil(math.sqrt(self.num_chars))

        # Calculate cell size to fit within resolution
        cell_size: int = min(self.resolution[0] // chars_per_row, self.resolution[1] // chars_per_col)

        # Calculate cell height
        cell_height: float = self.resolution[1] / chars_per_row

        return GridDimensions(characters_per_row=chars_per_row, characters_per_column=chars_per_col, cell_size=cell_size, cell_height=cell_height)

    def get_cell(self, index: int) -> GridCell:
        """Get the grid cell information for a given character index.

        Args:
            index: The index of the character in the grid

        Returns:
            GridCell containing position and coordinate information
        """
        # Calculate grid position
        grid_x: int = index % self.dimensions.characters_per_row
        grid_y: int = index // self.dimensions.characters_per_row

        # Calculate normalized coordinates
        norm_x1: float = grid_x / self.dimensions.characters_per_row
        norm_y1: float = (grid_y * self.dimensions.cell_height) / self.resolution[1]
        norm_x2: float = (grid_x + 1) / self.dimensions.characters_per_row
        norm_y2: float = ((grid_y + 1) * self.dimensions.cell_height) / self.resolution[1]

        # Convert to pixel coordinates
        pixel_x1: float = norm_x1 * self.resolution[0]
        pixel_y1: float = norm_y1 * self.resolution[1]
        pixel_x2: float = norm_x2 * self.resolution[0]
        pixel_y2: float = norm_y2 * self.resolution[1]

        return GridCell(
            x=grid_x,
            y=grid_y,
            pixel_x1=pixel_x1,
            pixel_y1=pixel_y1,
            pixel_x2=pixel_x2,
            pixel_y2=pixel_y2,
            norm_x1=norm_x1,
            norm_y1=1 - norm_y1,  # Invert Y-axis normalization
            norm_x2=norm_x2,
            norm_y2=1 - norm_y2,  # Invert Y-axis normalization
        )



"""Binary TLK (talk table) read/write: string entries, sound refs, and flags."""

from __future__ import annotations

import struct

from typing import TYPE_CHECKING

from pykotor.common.language import Language
from pykotor.common.misc import ResRef, WrappedInt
from pykotor.common.stream import ArrayHead
from pykotor.resource.formats.tlk.tlk_data import TLK
from pykotor.resource.type import ResourceReader, ResourceWriter, autoclose

if TYPE_CHECKING:
    from pykotor.resource.formats.tlk.tlk_data import TLKEntry
    from pykotor.resource.type import SOURCE_TYPES, TARGET_TYPES

_FILE_HEADER_SIZE = 20
_ENTRY_SIZE = 40


class TLKBinaryReader(ResourceReader):
    """Reads TLK (Talk Table) files.

    TLK files store localized strings used throughout the game for dialog, item descriptions,
    and other text content. Each entry can have text, sound references, and flags.

    References:
    ----------
        See tlk_data module docstring for engine addresses (K1 + TSL TODO). CTlkTable::CTlkTable (K1: 0x0041d8d0), CTlkTable::AddFile (K1: 0x0041d920), CTlkFile::CTlkFile (K1: 0x0041d810), "TLK " (K1: 0x0073ecb0), "tlk" extension (K1: 0x0074dd40).
        Missing Features:
        ----------------
        - ResRef lowercasing (reone lowercases sound resrefs)

    """

    def __init__(
        self,
        source: SOURCE_TYPES,
        offset: int = 0,
        size: int = 0,
        language: Language | None = None,
    ):
        super().__init__(source, offset, size)
        self._tlk: TLK
        self._texts_offset = 0
        self._text_headers: list[ArrayHead] = []
        self._language: Language | None = language

    @autoclose
    def load(self, *, auto_close: bool = True) -> TLK:  # noqa: FBT001, FBT002, ARG002
        self._tlk = TLK()
        self._texts_offset = 0
        self._text_headers = []

        self._reader.seek(0)

        self._load_file_header()
        # Optimize: batch read all entry headers at once instead of one-by-one
        self._load_entries_batch()
        for stringref, _entry in self._tlk:
            self._load_text(stringref)

        return self._tlk

    def _load_file_header(self):
        file_type = self._reader.read_string(4)
        file_version = self._reader.read_string(4)
        language_id = self._reader.read_uint32()
        string_count = self._reader.read_uint32()
        entries_offset = self._reader.read_uint32()

        if file_type != "TLK ":
            msg = "Invalid file type."
            raise ValueError(msg)
        if file_version != "V3.0":
            msg = "Invalid file version."
            raise ValueError(msg)

        self._tlk.language = Language(language_id) if self._language is None else self._language
        self._tlk.resize(string_count)

        self._texts_offset = entries_offset

    def _load_entries_batch(self):
        """Optimized batch loading of all entry headers.

        Reads all entry headers in one batch operation instead of one-by-one.
        This reduces seeks from N (one per entry) to 1 (one batch read).
        """
        string_count = len(self._tlk)
        if string_count == 0:
            return

        # Calculate where entries start (after 20-byte header)
        entries_start = _FILE_HEADER_SIZE
        # Each entry is 40 bytes: flags(4) + sound_resref(16) + volume(4) + pitch(4) + text_offset(4) + text_length(4) + sound_length(4)
        entries_size = string_count * _ENTRY_SIZE

        # Read all entry data at once
        self._reader.seek(entries_start)
        entries_data = self._reader.read_bytes(entries_size)

        # Parse entries from batch data
        for stringref in range(string_count):
            entry: TLKEntry = self._tlk.entries[stringref]
            offset = stringref * _ENTRY_SIZE

            # Parse entry: flags(4) + sound_resref(16) + volume(4) + pitch(4) + text_offset(4) + text_length(4) + sound_length(4)
            (entry_flags,) = struct.unpack("<I", entries_data[offset : offset + 4])
            sound_resref_bytes = entries_data[offset + 4 : offset + 20]
            # Find null terminator for sound_resref
            null_pos = sound_resref_bytes.find(b"\x00")
            if null_pos >= 0:
                sound_resref_bytes = sound_resref_bytes[:null_pos]
            sound_resref = sound_resref_bytes.decode("ascii", errors="ignore")
            _volume_variance, _pitch_variance, text_offset, text_length = struct.unpack("<IIII", entries_data[offset + 20 : offset + 36])
            (entry.sound_length,) = struct.unpack("<f", entries_data[offset + 36 : offset + 40])

            entry.text_present = (entry_flags & 0x0001) != 0
            entry.sound_present = (entry_flags & 0x0002) != 0
            entry.soundlength_present = (entry_flags & 0x0004) != 0
            entry.voiceover = ResRef(sound_resref)
            self._text_headers.append(ArrayHead(text_offset, text_length))

    def _load_entry(
        self,
        stringref: int,
    ):
        """Legacy method kept for compatibility - use _load_entries_batch instead."""
        entry: TLKEntry = self._tlk.entries[stringref]

        entry_flags = self._reader.read_uint32()

        # NOTE: reone lowercases sound_resref, PyKotor does not
        sound_resref = self._reader.read_string(16)
        _volume_variance = self._reader.read_uint32()  # unused
        _pitch_variance = self._reader.read_uint32()  # unused
        text_offset = self._reader.read_uint32()
        text_length = self._reader.read_uint32()
        entry.sound_length = self._reader.read_single()  # unused

        entry.text_present = (entry_flags & 0x0001) != 0  # Check if the TEXT_PRESENT flag is set
        entry.sound_present = (entry_flags & 0x0002) != 0  # Check if the SND_PRESENT flag is set
        entry.soundlength_present = (entry_flags & 0x0004) != 0  # Check if the SND_LENGTH flag is set
        entry.voiceover = ResRef(sound_resref)
        self._text_headers.append(ArrayHead(text_offset, text_length))

    def _load_text(
        self,
        stringref: int,
    ):
        text_header: ArrayHead = self._text_headers[stringref]

        self._reader.seek(text_header.offset + self._texts_offset)
        text: str = self._reader.read_string(text_header.length, encoding=self._tlk.language.get_encoding())

        self._tlk.entries[stringref].text = text


class TLKBinaryWriter(ResourceWriter):
    def __init__(
        self,
        tlk: TLK,
        target: TARGET_TYPES,
    ):
        super().__init__(target)
        self._tlk: TLK = tlk

    @autoclose
    def write(self, auto_close: bool = True):  # noqa: FBT001, FBT002, ARG002
        self._write_file_header()

        text_offset = WrappedInt(0)
        encoding: str | None = self._tlk.language.get_encoding()
        for entry in self._tlk.entries:
            self._write_entry(entry, text_offset)

        for entry in self._tlk.entries:
            self._writer.write_string(entry.text, encoding or "cp1252", errors="replace")

    def _calculate_entries_offset(self) -> int:
        return _FILE_HEADER_SIZE + len(self._tlk) * _ENTRY_SIZE

    def _write_file_header(self):
        language_id: int = self._tlk.language.value
        string_count: int = len(self._tlk)
        entries_offset: int = self._calculate_entries_offset()

        self._writer.write_string("TLK ", string_length=4)
        self._writer.write_string("V3.0", string_length=4)
        self._writer.write_uint32(language_id)
        self._writer.write_uint32(string_count)
        self._writer.write_uint32(entries_offset)

    def _write_entry(
        self,
        entry: TLKEntry,
        previous_offset: WrappedInt,
    ):
        sound_resref = str(entry.voiceover)
        text_offset = previous_offset.get()
        text_length = len(entry.text)

        entry_flags = 0  # Initialize entry_flags as zero
        if entry.text_present:
            entry_flags |= 0x0001  # TEXT_PRESENT: As we're writing text, let's assume it's always present
        if entry.sound_present:
            entry_flags |= 0x0002  # SND_PRESENT: If sound_resref is defined in this entry.
        if entry.soundlength_present:
            entry_flags |= 0x0004  # SND_LENGTH: Unused by KOTOR1 and 2. Determines whether the sound length field is utilized.

        self._writer.write_uint32(entry_flags)
        self._writer.write_string(sound_resref, string_length=16)
        self._writer.write_uint32(0)  # unused - volume variance
        self._writer.write_uint32(0)  # unused - pitch variance
        self._writer.write_uint32(text_offset)
        self._writer.write_uint32(text_length)
        self._writer.write_uint32(0)  # unused - sound length

        previous_offset += text_length
